// (c)2012-2013 by Louis Coder (Matthias Mueller).

// Contact: visual_basic_louis@yahoo.com; Visit www.louis-coder.com!
// This is (supposed to be) the implementation of a polynomial 2/3-SAT solver.
// 
// Additionally this Windows-console program allows to test that solver.
// The program starts with the procedure 'main()' at end of this file.
// The actual solver is located in the file 'ActualSolver.h' and its
// one and only interface procedure is 'CNF_Solve_Polynomial()'.

char Version[] = "v1.0";



int SAT_TYPE = 3; // 2 for 2-SAT or 3 for 3-SAT (can be changed at runtime by entering 's')



#define DIGIT_NUMBER_MAX						25		// greatest allowed DigitNumber for reading or creating (random) CNFs
#define CLAUSE_NUMBER_MAX						18500	// greatest allowed ClauseNumber for reading or creating (random) CNFs (tested maximum of 18400 for 3-SAT, TestMode 4 'TEST_MODE_COMPLETE_CNF', DigitNumber=25)

// #define MIN_OUTPUT // enable and re-compile to keep only basic output to speed up solving many (>1M) CNFs

// under Windows, the following suppresses warnings
// about unsafe charstring operations (gets() etc.)
#define _CRT_SECURE_NO_DEPRECATE

#include <tchar.h>		// _TCHAR
#include <stdio.h>		// fgets
#include <string.h>		// strlen
#include <windows.h>	// GetTickCount
#include <conio.h>		// _kbhit

#include <math.h>		// pow

// *** RANDOM ***
// NOTE: the Windows-function rand() is pretty slow as
// it determines some process id every time it is called.
// So we copied the original Windows-rand() implementation's
// code and built our own random generator out of it!

#ifndef RAND_MAX
#define RAND_MAX	32767
#endif

// IMPORTANT: this random number generator will always
// return the same number succession. That means when
// you re-start the solver and do exactly the same as
// at last start, also the very same CNFs will be
// created. So if the solver should ever fail, please
// write down how you used the solver (test mode,
// DigitNumber, ClauseNumber etc.) to reproduce the error.
// Besides that, it is not ingenious to re-start the
// solver and use the same DigitNumber/ClauseNumber etc.
// combination, as the same CNFs will be created and
// tested again (we already know that they work).

int RandValue = 0;

void SRand(int Seed)
{ // initializes the random generator

	// NOTE: as long as you pass the same value to this function
	// (or void its calling), the Rand() function will return the
	// same succession of ClauseNumbers. That means in every run of
	// this program, results are re-producible!

	RandValue = Seed;
}

int Rand()
{ // returns a 'random' number between (both including) 0 and 32767

	return (((RandValue = RandValue * 214013L
            + 2531011L) >> 16) & 0x7fff); // copied from original rand() code
}

// *** END OF RANDOM ***

// **************************************************************

void CNF_Show(int DigitNumber, int ClauseNumber, char* CNF)
{
	/*
	Example usage:

	CNF_Show(
		3,
		4,
		"000"	[NO comma - the compiler puts all sub strings together to one string!]
		"001"
		"010"
		"011");

	(or pass a char string as parameter CNF)
	*/

	for (int i = 0; i < ClauseNumber; i ++)
	{
		printf("   ");

		for (int j = 0; j < DigitNumber; j ++)
			printf("%c", CNF[i * DigitNumber + j]);

		printf("\n");
	}

	printf("\n");
}

// **************************************************************

void CNF_CreateRandomly(int DigitNumber, int ClauseNumber, char* CNF)
{
    for (int i = 0; i < ClauseNumber; i++)
    {
        for (int m = 0; m < DigitNumber; m ++)
			CNF[i * DigitNumber + m] = '-';

		for (int j = 0; j < SAT_TYPE; j++)
        {
        ReDo:
            int Pos = Rand() * DigitNumber / (RAND_MAX + 1); // 0...DigitNumber-1
            if (CNF[i * DigitNumber + Pos] != '-')
                goto ReDo;

			CNF[i * DigitNumber + Pos] = (Rand() < RAND_MAX / 2 ? '0' : '1');
        }
    }
}

bool CNF_ReadFromFile(int &DigitNumber, int &ClauseNumber, char* CNF, char* FilePath)
{ // will set DigitNumber and ClauseNumber; IGNORES META-DATA in file (literals and clauses are just read in as far as possible)!

    FILE*	F = NULL;
	char	Line[1024];

	printf("reading %s\n", FilePath);

	DigitNumber = 0;

	if (fopen_s(&F, FilePath, "r" ) == 0)
	{
		int ClauseIndex = 0;

		while (fgets(Line, 1024, F) > 0)
		{
			int Line_Length = (int)strlen(Line);

			if (Line_Length < 1)
				continue;

			if (Line[0] == 'c')
				continue;

			if (Line[0] == 'p')
				continue;

			if (Line[0] == '%')
				break;

			for (int m = 0; m < DIGIT_NUMBER_MAX; m ++)
				CNF[ClauseIndex * DIGIT_NUMBER_MAX + m] = '-';

			char Digits[16][8];

			for (int i = 0; i < 16; i ++)
				for (int m = 0; m < 8; m ++)
					Digits[i][m] = '\0';

			int LinePos = 0;
			int DigitPos = 0;

			for (int i = 0; i < 16; i ++)
			{
				for (; LinePos < Line_Length; LinePos ++)
					if (Line[LinePos] != ' ')
						break;
			
				DigitPos = 0; // the file can contain maximal 16 digits (without any terminating 0) per line
				for (; LinePos < Line_Length; LinePos ++)
					if (Line[LinePos] == ' ')
					{
						Digits[i][DigitPos ++] = '\0';
						break;
					}
					else
						if ((Digits[i][DigitPos ++] = Line[LinePos]) == '0' && DigitPos == 0)
							goto LineFinished;
			}
LineFinished:;

			for (int m = 0; m < 16; m ++)
			{
				int Val = 
					(Digits[m] == '\0' ? 0 : atoi(Digits[m])); // 1-based! Could be 0 if read 2-SAT file!

				if (Val > 0)
				{
					if (+Val <= DIGIT_NUMBER_MAX)
					{
						if (CNF[ClauseIndex * DIGIT_NUMBER_MAX + Val - 1] != '-')
							printf("error: digit already used (+)!");

						CNF[ClauseIndex * DIGIT_NUMBER_MAX + Val - 1] = '1';

						DigitNumber = max(DigitNumber, +Val);
					}
					else
						printf("error: digit out of bounds!");
				}
				if (Val < 0)
				{
					if (-Val <= DIGIT_NUMBER_MAX)
					{
						if (CNF[ClauseIndex * DIGIT_NUMBER_MAX - Val - 1] != '-')
							printf("error: digit already used (-)!");

						CNF[ClauseIndex * DIGIT_NUMBER_MAX - Val - 1] = '0';

						DigitNumber = max(DigitNumber, -Val);
					}
					else
						printf("error: digit out of bounds!");
				}
				/*
				if (Val == 0)
					printf("error: digit is 0!\n");
				*/

				if (SAT_TYPE == 2 && Val != 0 && m >= 2)
				{
					printf("warning: you must read files that contain EXACT-2-SAT clauses !!!\n");
					printf("         In the read file there are %d-SAT clauses !\n", m + 1);
					if (m + 1 == 3)
						printf("         You can press 's' in the main menu to switch to 3-SAT.");
					getchar();
					return false; // error
				}
				if (SAT_TYPE == 3 && Val != 0 && m >= 3)
				{
					printf("warning: you must read files that contain EXACT-3-SAT clauses !!!\n");
					printf("         In the read file there are %d-SAT clauses !\n", m + 1);
					getchar();
					return false; // error
				}
			}

			ClauseIndex ++;
		}

		printf("read %d lines.\n", ClauseIndex);

		ClauseNumber = ClauseIndex;

		if (ClauseNumber > CLAUSE_NUMBER_MAX)
		{
			printf("warning: in the read file, there are too many lines (clauses) !\n");
			getchar();
		}

		for (int m = 0; m < ClauseNumber; m ++)
			for (int n = 0; n < DigitNumber; n ++)
				CNF[m * DigitNumber + n] = CNF[m * DIGIT_NUMBER_MAX + n];
	}
	else
	{
		printf("error reading file !\n");

		ClauseNumber = 0;
	}

	if (F)
		fclose(F);

	return (ClauseNumber > 0);
}

void CNF_CreateRandomlyWithInvertedClauses(int DigitNumber, int ClauseNumber, char* CNF)
{
    for (int i = 0; i < ClauseNumber; i++)
    {
        for (int m = 0; m < DigitNumber; m ++)
			CNF[(i + 0) * DigitNumber + m] = '-';

		if (i + 1 < ClauseNumber)
			for (int m = 0; m < DigitNumber; m ++)
				CNF[(i + 1) * DigitNumber + m] = '-';

		for (int j = 0; j < SAT_TYPE; j++)
        {
        ReDo:
            int Pos = Rand() * DigitNumber / (RAND_MAX + 1); // 0...DigitNumber-1
            if (CNF[i * DigitNumber + Pos] != '-')
                goto ReDo;

			// add original clause
			CNF[(i + 0) * DigitNumber + Pos] = (Rand() < RAND_MAX / 2 ? '0' : '1');

			// add additionally the inverted clause
			if (i + 1 < ClauseNumber)
				CNF[(i + 1) * DigitNumber + Pos] = (CNF[(i + 0) * DigitNumber + Pos] == '0' ? '1' : '0'); // invert
        }

		i++;
    }
}

void CNF_CreateRandomlyWithAllPossibleClauses(int DigitNumber, int& ClauseNumber, char* CNF) // ClauseNumber and CNF are changed!
{ // creates a CNF that contains ALL possible clauses

	if (SAT_TYPE == 2)
	{
		ClauseNumber = 0;

		for (int Digit1 = 1; Digit1 <= DigitNumber - 2; Digit1 ++) // 3 digits per clause
			for (int Digit2 = Digit1 + 1; Digit2 <= DigitNumber - 1; Digit2 ++)
				for (int CIndex = 0; CIndex < 4; CIndex ++)
				{
					for (int i = 0; i < DigitNumber; i ++)
						CNF[ClauseNumber * DigitNumber + i] = '-';

					CNF[ClauseNumber * DigitNumber + (Digit1 - 1)] = (CIndex & 2 ? '1' : '0');
					CNF[ClauseNumber * DigitNumber + (Digit2 - 1)] = (CIndex & 1 ? '1' : '0');
					
					ClauseNumber ++;
				}

		// randomly swap clauses
		char* TempClause = new char[DigitNumber];

		for (int i = 0; i < ClauseNumber * 4; i ++)
		{
			// NOTE: result of multiplication could exceed int!
			int Index1 = (int)((int)Rand() * (int)ClauseNumber / (int)(RAND_MAX + 1));
			int Index2 = (int)((int)Rand() * (int)ClauseNumber / (int)(RAND_MAX + 1));

			for (int m = 0; m < DigitNumber; m ++)
				TempClause[m] = CNF[Index1 * DigitNumber + m];

			for (int m = 0; m < DigitNumber; m ++)
				CNF[Index1 * DigitNumber + m] = CNF[Index2 * DigitNumber + m];

			for (int m = 0; m < DigitNumber; m ++)
				CNF[Index2 * DigitNumber + m] = TempClause[m];
		}

		delete[] TempClause;
	}

	if (SAT_TYPE == 3)
	{
		ClauseNumber = 0;

		for (int Digit1 = 1; Digit1 <= DigitNumber - 2; Digit1 ++) // 3 digits per clause
			for (int Digit2 = Digit1 + 1; Digit2 <= DigitNumber - 1; Digit2 ++)
				for (int Digit3 = Digit2 + 1; Digit3 <= DigitNumber; Digit3 ++)
					for (int CIndex = 0; CIndex < 8; CIndex ++)
					{
						for (int i = 0; i < DigitNumber; i ++)
							CNF[ClauseNumber * DigitNumber + i] = '-';

						CNF[ClauseNumber * DigitNumber + (Digit1 - 1)] = (CIndex & 4 ? '1' : '0');
						CNF[ClauseNumber * DigitNumber + (Digit2 - 1)] = (CIndex & 2 ? '1' : '0');
						CNF[ClauseNumber * DigitNumber + (Digit3 - 1)] = (CIndex & 1 ? '1' : '0');
						
						ClauseNumber ++;
					}

		// randomly swap clauses
		char* TempClause = new char[DigitNumber];

		for (int i = 0; i < ClauseNumber * 4; i ++)
		{
			// NOTE: result of multiplication could exceed int!
			int Index1 = (int)((int)Rand() * (int)ClauseNumber / (int)(RAND_MAX + 1));
			int Index2 = (int)((int)Rand() * (int)ClauseNumber / (int)(RAND_MAX + 1));

			for (int m = 0; m < DigitNumber; m ++)
				TempClause[m] = CNF[Index1 * DigitNumber + m];

			for (int m = 0; m < DigitNumber; m ++)
				CNF[Index1 * DigitNumber + m] = CNF[Index2 * DigitNumber + m];

			for (int m = 0; m < DigitNumber; m ++)
				CNF[Index2 * DigitNumber + m] = TempClause[m];
		}

		delete[] TempClause;
	}
}

// **************************************************************

bool CNF_Solve_Exponential(int DigitNumber, int ClauseNumber, char* CNF)
{ // notice that you should NOT USE MORE than 25-30 digits or solving will take very much time!

	// Notice that this solver function may try out
	// ALL possible solutions (if CNF is unsat)
	// and thus may need O(2^DigitNumber) time.

    char	Solution[100];
	int		SolutionNumber = 0;
	bool	Solvable = false;

    for (int i = 0; i < DigitNumber; i++)
        Solution[i] = '0';

    while (true)
    {
		// check if ALL clauses of the CNF are satisfied by the current solution:
        for (int ClauseFor = 0; ClauseFor < ClauseNumber; ClauseFor ++)
        {
			bool ClauseSatisfied = false;

            for (int j = 0; j < DigitNumber; j++)
			{
                if (Solution[j] == CNF[ClauseFor * DigitNumber + j])
				{
					ClauseSatisfied = true;
					break;
				}
			}

            if (ClauseSatisfied == false)
				goto NextSolution;
        }

		// if we arrived here, we found a solution
		SolutionNumber ++;

		// show the first 5 solutions (if any):
		if (SolutionNumber == 5)
		{
#ifndef MIN_OUTPUT
			printf("[...]\n\n");
#endif
			return true;
		}
		else
		if (SolutionNumber <= 4)
		{
#ifndef MIN_OUTPUT
			printf("(fail-safe) solution:\n");

			for (int j = 0; j < DigitNumber; j ++)
				if (Solution[j] == '0')
					printf("0");
				else
				if (Solution[j] == '1')
					printf("1");

			printf("\n");
#endif
			Solvable = true;
		}

NextSolution: ;
		// get next possible solution...
        for (int j = DigitNumber - 1; j >= 0; j--)
            if (Solution[j] == '0')
            {
                Solution[j] = '1';
                goto TestSolution;
            }
            else
                Solution[j] = '0';

		if (Solvable)
		{
#ifndef MIN_OUTPUT
			printf("\n");
#endif
			return true;
		}
		else
		{
			return false;
		}

TestSolution: ;
		// ...and check (in the next while-iteration) if it satisfies the CNF
    }
}

// **************************************************************

#define TEST_MODE_OTHER					-1
#define TEST_MODE_EXIT					0
#define TEST_MODE_RANDOM_CNF			1
#define TEST_MODE_FILES					2
#define TEST_MODE_RANDOM_INVERTED_CNF	3
#define TEST_MODE_COMPLETE_CNF			4
#define TEST_MODE_PIGEON_2				5
#define TEST_MODE_PIGEON_3				6
#define TEST_MODE_ALL_COMPLETE			7

#include "ActualSolver.h" // the C++ compiler will insert here the file containing the actual polynomial solver!

// FOR TEST_MODE_ALL_COMPLETE >>>

int				PossibleClausesOnOffCombIndex;

// <<< END OF FOR TEST_MODE_ALL_COMPLETE

int _tmain(int argc, _TCHAR* argv[]) // the program starts here
{
	static char	CNF[CLAUSE_NUMBER_MAX * DIGIT_NUMBER_MAX]; // possible digits: 0, 1, -
	// 
	// NOTE: we use the 'CNF Table'-format, see documentation.
	// Clauses appear in a coherent string, one clause after another.
	// CNF[ClauseIndex * DigitNumber + DigitIndex] is the DigitIndex-st char
	// in the ClauseIndex-st clause.

	while (true)
	{
		int		TestMode;
		int		TestRunNumber;
		int		TestRunStartIndex;

		int		DN; // digit number
		int		CN; // clause number

		bool	FileSolvable;
		char*	FilePathFormat;

		bool	ShowCNFFlag;

		char	S[1024]; // stores any user input from gets()

		while (true)
		{
			// SET STANDARD TEST (just hit Return to run) >>>:

			TestMode = TEST_MODE_RANDOM_INVERTED_CNF;
			TestRunNumber = 100000000;
			TestRunStartIndex = 1;

			DN = 10;

			if (SAT_TYPE == 2)
			{
				CN = 20;
			}
			if (SAT_TYPE == 3)
			{
				CN = 50;
			}

#define		TEST_RUN_INDEX_START	0 // set to the TestRun index that caused an error (should not happen at all!)

			FileSolvable = false;
			FilePathFormat = "";

			ShowCNFFlag = false;

			// <<< END OF SETTING STANDARD TEST

			printf("**********************************************************************\n");
			printf("POLYNOMIAL EXACT-2/3-SAT SOLVER DEMO %s\n", Version);
			printf("    from: www.louis-coder.com\n");
			printf("\n");
			printf("Please read documentation for details.\n");
			printf("\n");
			if (SAT_TYPE == 2)
			{
				printf("TWO-SAT: Works as 2-SAT solver.\n");
				printf("Press 's' to switch to 3-SAT.\n");
			}
			if (SAT_TYPE == 3)
			{
				printf("THREE-SAT: Works as 3-SAT solver.\n");
				printf("Press 's' to switch to 2-SAT.\n");
			}
			printf("\n");
			printf("Please select testing mode (checks if polynomial = exponential result):\n");
			printf("\n");
			printf("1: test random CNFs\n");
			printf("2: test CNF from file\n");
			printf("3: test random CNFs with each clause additionally negated\n");
			printf("4: test complete CNF\n");
			printf("5: test pigeon hole problem n=2, converted to pure 3-SAT\n");
			printf("6: test pigeon hole problem n=3, converted to pure 3-SAT\n");
			printf("7: test all CNFs possible\n");
			printf("\n");
			printf("e: exit\n");
			printf("\n");
			printf("Just hit Enter to run standard test\n");
			printf("(random CNFs, DigitNumber=%d, ClauseNumber = %d, %d test runs)\n", DN, CN, TestRunNumber);
			
			gets(S); // ignore any compiler warnings like 'gets() is not safe', we don't need a bullet-proof input method!
			
			switch (S[0])
			{
			case 'e':
				TestMode = TEST_MODE_EXIT;
				break;

			case 's':
				if (SAT_TYPE == 2)
				{
					SAT_TYPE = 3;
					printf("Switched to THREE-SAT.\n");
					Sleep(3000);
					printf("\n");
					continue;
				}
				if (SAT_TYPE == 3)
				{
					SAT_TYPE = 2;
					printf("Switched to TWO-SAT.\n");
					Sleep(3000);
					printf("\n");
					continue;
				}
				continue; // should not happen

			case '1':
				printf("**************************************************************\n");
				printf("* The solver will create random CNFs which are solved by     *\n");
				printf("* once the exponential 'brute-force' algorithm and also by   *\n");
				printf("* the polynomial one. The results are checked for equality.  *\n");
				printf("*                                                            *\n");
				printf("* Tip: do not use a DigitNumber much greater than 10 as the  *\n");
				printf("* polynomial solving has a complexity of O(DigitNumber^10)   *\n");
				printf("* and is therefore not very fast on common PCs.              *\n");
				printf("*                                                            *\n");
				printf("* Notice: the solver tells only *if* the CNF is solvable,    *\n");
				printf("* it does not find any explicit solution(s).                 *\n");
				printf("**************************************************************\n");
				printf("\n");
				if (SAT_TYPE == 2)
				{
					printf("DigitNumber       (e.g. 12): ");
					gets(S);
					DN = atoi(S);
					printf("ClauseNumber      (e.g. 24): ");
				}
				if (SAT_TYPE == 3)
				{
					printf("DigitNumber        (e.g. 8): ");
					gets(S);
					DN = atoi(S);
					printf("ClauseNumber      (e.g. 40): ");
				}
				gets(S);
				CN = atoi(S);
				printf("TestRunNumber (e.g. 100000): ");
				gets(S);
				TestRunNumber = atoi(S);
				TestRunStartIndex = 1;
				printf("Show CNFs? y/n     (e.g. n): ");
				gets(S);
				ShowCNFFlag = (S[0] == 'y');

				TestMode = TEST_MODE_RANDOM_CNF;
				break;

			case '2':
				printf("*******************************************************************************\n");
				printf("* The solver is able to read CNFs from files that have the DIMACS format.     *\n");
				printf("* See 'http://www.cs.ubc.ca/~hoos/SATLIB/Benchmarks/SAT/satformat.ps'.        *\n");
				printf("*                                                                             *\n");
				printf("* You can enter the relative or absolute path to the file that is to be read. *\n");
				printf("*                                                                             *\n");
				printf("* Please notice that the solver reads as many clauses as there are in         *\n");
				printf("* the file, any meta data or/and header is ignored.                           *\n");
				printf("*                                                                             *\n");
				printf("* Also notice that the solver tells only *if* the CNF is solvable,            *\n");
				printf("* it does not find any explicit solution(s).                                  *\n");
				printf("*******************************************************************************\n");
				printf("\n");
				printf("Path            : ");
				gets(S);
				FilePathFormat = _strdup(S);
				DN = 10; // will be changed when reading file
				CN = 10; // will be changed when reading file
				TestRunNumber = 1;
				TestRunStartIndex = 1;

				TestMode = TEST_MODE_FILES;
				break;

			case '3':
				printf("**************************************************************\n");
				printf("* The solver will create random CNFs which are solved by     *\n");
				printf("* once the exponential 'brute-force' algorithm and also by   *\n");
				printf("* the polynomial one. The results are checked for equality.  *\n");
				printf("*                                                            *\n");
				printf("* In the random CNF, each clause appears also negated.       *\n");
				printf("* I realized that this lead to more errors in early          *\n");
				printf("* versions of the polynomial solver, so I count that         *\n");
				printf("* kind of CNFs as 'harder to solve'.                         *\n");
				printf("*                                                            *\n");
				printf("* Tip: do not use a DigitNumber much greater than 10 as the  *\n");
				printf("* polynomial solving has a complexity of O(DigitNumber^10)   *\n");
				printf("* and is therefore not very fast on common PCs.              *\n");
				printf("*                                                            *\n");
				printf("* Notice: the solver tells only *if* the CNF is solvable,    *\n");
				printf("* it does not find any explicit solution(s).                 *\n");
				printf("**************************************************************\n");
				printf("\n");
				if (SAT_TYPE == 2)
				{
					printf("DigitNumber       (e.g. 12): ");
					gets(S);
					DN = atoi(S);
					printf("ClauseNumber      (e.g. 24): ");
				}
				if (SAT_TYPE == 3)
				{
					printf("DigitNumber        (e.g. 8): ");
					gets(S);
					DN = atoi(S);
					printf("ClauseNumber      (e.g. 40): ");
				}
				gets(S);
				CN = atoi(S);
				printf("TestRunNumber (e.g. 100000): ");
				gets(S);
				TestRunNumber = atoi(S);
				TestRunStartIndex = 1;
				printf("Show CNFs? y/n     (e.g. n): ");
				gets(S);
				ShowCNFFlag = (S[0] == 'y');

				TestMode = TEST_MODE_RANDOM_INVERTED_CNF;
				break;

			case '4':
				printf("**************************************************************\n");
				printf("* The solver will create random CNFs which are solved by     *\n");
				printf("* once the exponential 'brute-force' algorithm and also by   *\n");
				printf("* the polynomial one. The results are checked for equality.  *\n");
				printf("*                                                            *\n");
				printf("* The created CNFs will consists of *all* possible clauses a *\n");
				printf("* CNF with the entered dimensions could have. Those CNFs are *\n");
				printf("* always unsatisfiable, that means they do not have any      *\n");
				printf("* solution(s). The CNFs' clauses are permutated randomly.    *\n");
				printf("*                                                            *\n");
				printf("* Tip: do not use a DigitNumber much greater than 10 as the  *\n");
				printf("* polynomial solving has a complexity of O(DigitNumber^10)   *\n");
				printf("* and is therefore not very fast on common PCs.              *\n");
				printf("*                                                            *\n");
				printf("* Notice: the solver tells only *if* the CNF is solvable,    *\n");
				printf("* it does not find any explicit solution(s).                 *\n");
				printf("**************************************************************\n");
				printf("\n");
				printf("DigitNumber       (e.g. 10): ");
				gets(S);
				DN = atoi(S);
				CN = 1; // will be set later
				printf("TestRunNumber (e.g. 100000): ");
				gets(S);
				TestRunNumber = atoi(S);
				TestRunStartIndex = 1;
				printf("Show CNFs? y/n     (e.g. n): ");
				gets(S);
				ShowCNFFlag = (S[0] == 'y');

				if (DN > DIGIT_NUMBER_MAX) // verify
					DN = 0; // will lead to an error message

				TestMode = TEST_MODE_COMPLETE_CNF;
				break;

			case '5':

				if (SAT_TYPE == 2)
				{
					SAT_TYPE = 3;
					printf("Switched to THREE-SAT.\n");
					Sleep(3000);
					printf("\n");
				}

				if (SAT_TYPE == 3)
				{
					printf("*******************************************************************************\n");
					printf("* The pigeon hole problem is a kind of CNF that makes any resolution-based    *\n");
					printf("* solver need exponential time for solving.                                   *\n");
					printf("*                                                                             *\n");
					printf("* As (among other things) the polynomial solver you are currently using is    *\n");
					printf("* not a resolution-based solver, it can detect the pigeon hole instances as   *\n");
					printf("* unsatisfiable within polynomial time and space.                             *\n");
					printf("*                                                                             *\n");
					printf("* Unfortunately this polynomial solver cannot solve large pigeon hole         *\n");
					printf("* problems within a reasonable time on common PCs as the solver generally     *\n");
					printf("* needs O(n^10) time. Therefore in the following test a rather small          *\n");
					printf("* pigeon hole problem instance is solved for demonstration purposes.          *\n");
					printf("*******************************************************************************\n");
					printf("\n");
					printf("Press Return to begin test...\n");
					printf("\n");

					getchar();

					TestMode = TEST_MODE_PIGEON_2;
					TestRunNumber = 1;
					TestRunStartIndex = 1;

					DN = 7;
					CN = 12;

					ShowCNFFlag = true;
				}

				break;

			case '6':

				if (SAT_TYPE == 2)
				{
					SAT_TYPE = 3;
					printf("Switched to THREE-SAT.\n");
					Sleep(3000);
					printf("\n");
				}

				if (SAT_TYPE == 3)
				{
					printf("*******************************************************************************\n");
					printf("* The pigeon hole problem is a kind of CNF that makes any resolution-based    *\n");
					printf("* solver need exponential time for solving.                                   *\n");
					printf("*                                                                             *\n");
					printf("* As (among other things) the polynomial solver you are currently using is    *\n");
					printf("* not a resolution-based solver, it can detect the pigeon hole instances as   *\n");
					printf("* unsatisfiable within polynomial time and space.                             *\n");
					printf("*                                                                             *\n");
					printf("* Unfortunately this polynomial solver cannot solve large pigeon hole         *\n");
					printf("* problems within a reasonable time on common PCs as the solver generally     *\n");
					printf("* needs O(n^10) time. Therefore in the following test a rather small          *\n");
					printf("* pigeon hole problem instance is solved for demonstration purposes.          *\n");
					printf("*******************************************************************************\n");
					printf("\n");
					printf("Press Return to begin test...\n");
					printf("\n");

					getchar();

					TestMode = TEST_MODE_PIGEON_3;
					TestRunNumber = 1;
					TestRunStartIndex = 1;

					DN = 13;
					CN = 26;

					ShowCNFFlag = true;
				}

				break;

			case '7':

				do
				{
					printf("**************************************************************\n");
					printf("* The solver will solve *all* possible CNFs that have the    *\n");
					printf("* DigitNumber you entered. The DigitNumber is the literal    *\n");
					printf("* index range within the CNF.                                *\n");
					printf("*                                                            *\n");
					printf("* Notice that it can take a long time to create and check    *\n");
					printf("* all possible CNFs if the DigitNumber is large.             *\n");
					printf("*                                                            *\n");
					printf("* Because of this, there are the following limits:           *\n");
					printf("*                                                            *\n");
					printf("* 2-SAT: DigitNumber = 3 or 4,                               *\n");
					printf("* 3-SAT: DigitNumber = 3.                                    *\n");
					printf("**************************************************************\n");
					printf("\n");
					printf("DigitNumber        (e.g. 3): ");
					gets(S);
					printf("\n");

					DN = atoi(S);
					CN = 1; // will be set later
				}
				while (!(
					(SAT_TYPE == 2 && DN >= 3 && DN <= 4) ||
					(SAT_TYPE == 3 && DN >= 3 && DN <= 3)));

				TestMode = TEST_MODE_ALL_COMPLETE;

				if (SAT_TYPE == 2 && DN == 3)
					TestRunNumber = (int)pow(2.0, 4 * 3); // PossibleClauseNumber = 4 0/1 Combinations * 3 ClauseTableColumns...
				if (SAT_TYPE == 2 && DN == 4)
					TestRunNumber = (int)pow(2.0, 4 * 6); // ...and there are 2^PossibleClauseNumber on/off combinations of PossibleClauses
				if (SAT_TYPE == 3 && DN == 3)
					TestRunNumber = (int)pow(2.0, 8 * 1);

				TestRunStartIndex = 1;

				ShowCNFFlag = true;

				break;
			}

			if (TestMode == TEST_MODE_EXIT || (
				TestMode >= TEST_MODE_RANDOM_CNF && TestMode <= TEST_MODE_ALL_COMPLETE &&
				DN >= 3 && DN <= DIGIT_NUMBER_MAX &&
				CN >= 1 && CN <= CLAUSE_NUMBER_MAX &&
				TestRunNumber >= 1 &&
				TestRunNumber <= 100000000 &&
				TestRunStartIndex >= 1 &&
				TestRunStartIndex <= 100000000))
			{
				printf("\n");
				break;
			}
			else
			{
				printf("\n");
				printf("ERROR: Wrong input !\n");
				printf("       Probably you tried to exceed the CNF size limits.\n");
				printf("       Try again with other parameters.\n");
				Sleep(3000);
				printf("\n");
			}
		}

		if (TestMode == TEST_MODE_EXIT)
			break;

		int		SuccessNumber = 0;
		int		ErrorNumber = 0;
		int		UnknownNumber = 0;

		int		SolvableNumber = 0;
		int		UNSATNumber = 0;

		printf("\n");

		// NOTE: when you begin test runs, the random number generator is initialized each time the
		// test loop is entered. This means that if you use the same DigitNumber and ClauseNumber,
		// there will ALWAYS the same formulae created, no matter what test runs you did before!
		// 
		// So you do NOT need to restart the program after each test loop to be able to create
		// re-producible tests.

		SRand(1981); // same SRand() parameter means same value succession of Rand() everytime you run the program

		for (int TestRunIndex = TestRunStartIndex; TestRunIndex < TestRunStartIndex + TestRunNumber; TestRunIndex += 1)
		{
			if (TestRunIndex >= TEST_RUN_INDEX_START) // for debugging
				printf("\n*********************************************************** TestRun #%d\n", TestRunIndex);

			/* // the following varies the CNF dimensions permanently in a hard-coded range:
			// **************************************************
			if (SAT_TYPE == 2)
			{
				DN = 3 + Rand() * 15 / RAND_MAX;
				CN = 2 * DN * 1 / 4 + Rand() * 2 * DN * 7 / 4 / RAND_MAX;
			}
			if (SAT_TYPE == 3)
			{
				DN = 3 + Rand() * 6 / RAND_MAX;
				CN = 5 * DN * 1 / 4 + Rand() * 5 * DN * 7 / 4 / RAND_MAX;
			}
			// **************************************************
			*/

			bool SolvableExponential;
			bool SolvablePolynomial;

			switch (TestMode)
			{
			case TEST_MODE_RANDOM_CNF:

				CNF_CreateRandomly(DN, CN, CNF); // uses the random number generator
				break;

			case TEST_MODE_FILES:

				char FilePath[1000];

				sprintf(FilePath, (const char*)FilePathFormat, TestRunIndex);

				if (CNF_ReadFromFile(DN, CN, CNF, FilePath) == false) // does NOT use the random number generator
					TestMode = TEST_MODE_EXIT;

				break;

			case TEST_MODE_RANDOM_INVERTED_CNF:

				CNF_CreateRandomlyWithInvertedClauses(DN, CN, CNF); // uses the random number generator
				break;

			case TEST_MODE_COMPLETE_CNF:

				CNF_CreateRandomlyWithAllPossibleClauses(DN, CN, CNF); // uses the random number generator
				break;

			case TEST_MODE_PIGEON_2:

				DN = 7;
				CN = 13;

				memcpy(CNF,
					"11----1"
					"--11--1"
					"----111"
					"0-0---1"
					"0---0-1"
					"--0-0-1"
					"-0-0--1"
					"-0---01"
					"---0-01"
					"11----0"
					"01----0"
					"10----0"
					"00----0",
					DN * CN * sizeof(char));

				break;

			case TEST_MODE_PIGEON_3:

				DN = 13;
				CN = 26;

				memcpy(CNF,
					"111----------"
					"---111-------"
					"------111----"
					"---------111-"
					"0--0--------1"
					"0-----0-----1"
					"0--------0--1"
					"---0--0-----1"
					"---0-----0--1"
					"------0--0--1"
					"-0--0-------1"
					"-0-----0----1"
					"-0--------0-1"
					"----0--0----1"
					"----0-----0-1"
					"-------0--0-1"
					"--0--0------1"
					"--0-----0---1"
					"--0--------01"
					"-----0--0---1"
					"-----0-----01"
					"--------0--01"
					"11----------0"
					"10----------0"
					"01----------0"
					"00----------0",
					DN * CN * sizeof(char));

				break;

			case TEST_MODE_ALL_COMPLETE:

				DigitNumber = DN;
				ClauseNumber = CN;

				if (DigitNumberOld != DigitNumber ||
					SAT_TYPE_OLD != SAT_TYPE) // at first run or when DigitNumber or/and SAT_TYPE changed
				{
					PossibleClauses_Create(); // initialize PossibleClauses[][] and PossibleClauseNumber
					PossibleClausesOnOffCombIndex = 1;
				}
				else
				{
					PossibleClausesOnOffCombIndex ++;
					if (PossibleClausesOnOffCombIndex > pow(2.0, PossibleClauseNumber))
						TestMode = TEST_MODE_EXIT;
				}

				ClauseNumber = 0;

				// IMPORTANT NOTE: the following works only if
				// PossibleClauseNumber < 32!
				// Maybe use u64 for PossibleClausesOnOffCombIndex.

				for (int i = 0; i < PossibleClauseNumber; i ++)
				{
					if (PossibleClausesOnOffCombIndex & (1 << i))
					{
						for (int m = 0; m < DN; m ++)
							CNF[ClauseNumber * DigitNumber + m] = PossibleClauses[i][m];

						ClauseNumber ++;
					}
				}

				CN = ClauseNumber;
				break;

			default:

				printf("internal error !\n");
				break;
			}

			if (TestRunIndex < TEST_RUN_INDEX_START)
				continue;

			// if a file could not be read:
			if (TestMode == TEST_MODE_EXIT)
				break;
			
			// show CNF info
			printf("DigitNumber=%d x ClauseNumber=%d; TestMode=%d, TestRunNumber=%d, %d-SAT\n", DN, CN, TestMode, TestRunNumber, SAT_TYPE);
			printf("\n");

			// show CNF if desired
#ifndef MIN_OUTPUT
			if (ShowCNFFlag)
			{
				printf("CNF to be solved:\n");
				CNF_Show(DN, CN, CNF);
			}
#endif
			// *******************
			// *** SOLVING >>> ***
			// *******************

			if (DN <= 25) // more does quickly get too slow
			{
#ifndef MIN_OUTPUT
				printf("\n");
				printf("1) Using exponential algorithm to check for solvability...\n");
				printf("\n");

				if (SolvableExponential = CNF_Solve_Exponential(DN, CN, CNF)) // the exponential (O(2^n)) algorithm (guaranteed to always work)
					printf(" * Exponential solution: solvable          (! for sure)\n");
				else
					printf(" * Exponential solution: UNSAT             (! for sure)\n");
#else
				SolvableExponential = CNF_Solve_Exponential(DN, CN, CNF);
#endif
			}

			DWORD t = GetTickCount();

			bool PolynomialSolvingAbortedFlag = false;
Again:;
#ifndef MIN_OUTPUT
			printf("\n");
			printf("2) Using polynomial algorithm to check for solvability...\n");
			printf("\n");
#endif
			if (SolvablePolynomial = CNF_Solve_Polynomial(DN, CN, CNF, PolynomialSolvingAbortedFlag)) // the polynomial algorithm (supposed to work ;-P )
			{
#ifndef MIN_OUTPUT
				printf("\n");
				t = (GetTickCount() - t + 500) / 1000; // +500 ms to round correctly
				printf(" * Polynomial solution : solvable          (? to be verified) (%d sec.)\n", t);
#endif
				SolvableNumber ++;
			}
			else
			if (PolynomialSolvingAbortedFlag)
			{
				printf("\n");
				t = (GetTickCount() - t + 500) / 1000; // +500 ms to round correctly
				printf(" * Polynomial solution : ABORTED           (? solvability unknown) (%d sec.)\n", t);
			}
			else
			{
#ifndef MIN_OUTPUT
				printf("\n");
				t = (GetTickCount() - t + 500) / 1000; // +500 ms to round correctly
				printf(" * Polynomial solution : UNSAT             (? to be verified) (%d sec.)\n", t);
#endif
				UNSATNumber ++;
			}

			if (PolynomialSolvingAbortedFlag)
			{
				printf("\n=> ????? :/\n\n", t);

				UnknownNumber ++;
			}
			else if (DN > 25 /* && TestMode != TEST_MODE_FILES */ && TestMode != TEST_MODE_COMPLETE_CNF)
			{
				printf("\n=> ????? :/\n\n", t);

				UnknownNumber ++;
			}
			else if (SolvableExponential == SolvablePolynomial)
			{
				printf("\n=> Polynomial algorithm worked :)\n\n", t);

				SuccessNumber ++;
			}
			else
			{
				printf("\n=> Polynomial algorithm possibly failed :(\n\n", t);

				getchar();

				ErrorNumber ++;

				goto Again;

				// break;
			}

			// **************************
			// *** <<< END OF SOLVING ***
			// **************************

			if (_kbhit() != 0)
			{
				printf("abort? (y/n) ");
				gets(S);
				if (S[0] == 'y')
					break;
				else
					printf("\n");
			}
		}

		printf("\n");
		printf("++++++++++++++++++++\n");
		printf("+++ End of Test. +++\n");
		printf("++++++++++++++++++++\n");
		printf("\n");
		printf("Success=%d (%d Solvable, %d UNSAT)   Error=%d   Unknown=%d\n", SuccessNumber, SolvableNumber, UNSATNumber, ErrorNumber, UnknownNumber);

		if (SuccessNumber > 0 && ErrorNumber == 0 && UnknownNumber == 0)
			printf("\n:) :) :) :) :)\n");

		printf("\n");

		printf("Press Return to finish test...\n");
		getchar();
	}

	return 0;
}
