// (c)2013-2015 by Louis Coder.

// ********************************** RANDOM **********************************

// NOTE: the Windows-function rand() is rather slow as
// it determines some process id every time it is called.
// So we copied the original Windows-rand() implementation's
// code and built our own random generator out of it!

#define RAND_MAX_SELF_DEF	(32767) // attention: on (some?) Linux systems, the pre-defined RAND_MAX is INT_MAX!

#define RAND_MAX_PLUS_ONE	(RAND_MAX_SELF_DEF + 1)

int RandValue = 0;

void SRand(int Seed)
{ // initializes the random generator

	// NOTE: as long as you pass the same value to this function
	// (or void its calling), the Rand() function will return the
	// same succession of ClauseNumbers. That means in every run of
	// this program, results are re-producible, as long as you
	// pass the same initialization value to SRand() and SRand2().
	// You can do this by setting USE_SEED X in User_Interface.cpp
	// to any constant X!

	RandValue = Seed;
}

int Rand()
{ // returns a 'random' number between (both including) 0 and 32767

	return (((RandValue = RandValue * 214013L
            + 2531011L) >> 16) & 0x7fff); // copied from original rand() code
}

// NOTE: we implement a second random number generator that
// is to be used by code that does NOT create random CNFs;
// The second random generator is some kind of 'helping
// random generator'.

int Rand2Value = 0;

void SRand2(int Seed2)
{ // initializes the random generator

	// NOTE: as long as you pass the same value to this function
	// (or void its calling), the Rand2() function will return the
	// same succession of ClauseNumbers. That means in every run of
	// this program, results are re-producible, as long as you
	// pass the same initialization value to SRand() and SRand2().
	// You can do this by setting USE_SEED X in User_Interface.cpp
	// to any constant X!

	Rand2Value = Seed2;
}

int Rand2()
{ // returns a 'random' number between (both including) 0 and 32767

	return (((Rand2Value = Rand2Value * 214013L
            + 2531011L) >> 16) & 0x7fff); // copied from original rand() code
}

// ****************************** END OF RANDOM *******************************
// ******************************* DEBUG OUTPUT *******************************

void Clause_Show(const int DigitNumber, const char* Clause, const bool NewLine = true)
{
	// This procedure is thought to be used for debugging.
	// You can call it like this:
	// Clause_Show(DigitNumber, PossibleClauses[x]);
	// or
	// Clause_Show(DigitNumber, &CNF[ClauseIndex * DigitNumber]);.

	if (Clause == NULL)
		return;

	for (int m = 0; m < DigitNumber; m ++)
		printf("%c", Clause[m]);
	if (NewLine)
		printf("\n");
}

void CNF_Show(const int DigitNumber, const int ClauseNumber, const char* CNF)
{
	/*
	Example usage:

	CNF_Show(
		3,
		4,
		"000"	[NO comma - the compiler puts all sub strings together to one big string!]
		"001"
		"010"
		"011");

	(or pass a char string as parameter CNF)
	*/

	// printf("CNF:\n");
	for (int i = 0; i < ClauseNumber; i ++)
	{
		printf("    ");
		Clause_Show(DigitNumber, &CNF[i * DigitNumber], true);
	}
	printf("\n");
}

// *************************** END OF DEBUG OUTPUT ****************************
// *************************** SAT FORMULA CREATION ***************************

void CNF_CreateRandomly(const int DigitNumber, const int ClauseNumber, char* CNF) // CNF is changed
{
	for (int i = 0; i < ClauseNumber; i++)
	{
		for (int m = 0; m < DigitNumber; m ++)
			CNF[i * DigitNumber + m] = '-';

		for (int j = 0; j < SAT_TYPE; j ++)
		{
			int Pos;
			while (true)
			{
				Pos = Rand() * DigitNumber / RAND_MAX_PLUS_ONE; // 0...DigitNumber-1
				if (CNF[i * DigitNumber + Pos] == '-')
					break;
			}
			CNF[i * DigitNumber + Pos] = (Rand() < RAND_MAX_SELF_DEF / 2 ? '0' : '1');
		}
	}
}

bool CNF_ReadFromFile(int &DigitNumber, int &ClauseNumber, char* CNF, const char* FilePath) // DigitNumber, ClauseNumber and CNF is changed
{ // will set DigitNumber and ClauseNumber; IGNORES META-DATA in file (literals and clauses are just read in as many as possible)!

	FILE*	F = NULL;
	char	Line[1024];

	printf("reading %s\n", FilePath);

	DigitNumber = 0;

#ifdef _WIN32
	if (fopen_s(&F, FilePath, "r") == 0)
#else
	if ((F = fopen(FilePath, "r")) != NULL)
#endif
	{
		int ClauseIndex = 0;

		while (fgets(Line, 1024, F) > 0)
		{
			int Line_Length = (int)strlen(Line); // attention: can't detect one '\n' char as zero-length right!

			if (Line_Length < 1 ||
				Line[0] == '\n' ||
				Line[0] == '\r')
				continue;

			if (Line[0] == 'c')
				continue;

			if (Line[0] == 'p')
				continue;

			if (Line[0] == '%')
				break;

			for (int m = 0; m < DIGIT_NUMBER_MAX; m ++)
				CNF[ClauseIndex * DIGIT_NUMBER_MAX + m] = '-';

			char Digits[16][8];

			for (int i = 0; i < 16; i ++)
				for (int m = 0; m < 8; m ++)
					Digits[i][m] = '\0';

			bool	LineFinished = false;
			int		LinePos = 0;
			int		DigitPos = 0;

			for (int i = 0; i < 16; i ++)
			{
				for (; LinePos < Line_Length; LinePos ++)
					if (Line[LinePos] != ' ')
						break;

				DigitPos = 0; // the file can contain maximal 16 digits (without any terminating 0) per line
				for (; LinePos < Line_Length; LinePos ++)
				{
					if (Line[LinePos] == ' ')
					{
						Digits[i][DigitPos ++] = '\0';
						break;
					}
					else
					{
						if ((Digits[i][DigitPos ++] = Line[LinePos]) == '0' && DigitPos == 0)
						{
							LineFinished = true;
							break;
						}
					}
					if (LineFinished)
						break;
				}
			}

			for (int m = 0; m < 16; m ++)
			{
				int Val =
					(Digits[m][0] == '\0' ? 0 : atoi(Digits[m])); // 1-based! Could be 0 if read 2-SAT file!

				if (Val > 0)
				{
					if (+Val <= DIGIT_NUMBER_MAX)
					{
						if (CNF[ClauseIndex * DIGIT_NUMBER_MAX + Val - 1] != '-')
							printf("error: digit already used (+)!");

						CNF[ClauseIndex * DIGIT_NUMBER_MAX + Val - 1] = '1';

						DigitNumber = MAX(DigitNumber, +Val);
					}
					else
						printf("error: digit out of bounds!");
				}
				if (Val < 0)
				{
					if (-Val <= DIGIT_NUMBER_MAX)
					{
						if (CNF[ClauseIndex * DIGIT_NUMBER_MAX - Val - 1] != '-')
							printf("error: digit already used (-)!");

						CNF[ClauseIndex * DIGIT_NUMBER_MAX - Val - 1] = '0';

						DigitNumber = MAX(DigitNumber, -Val);
					}
					else
						printf("error: digit out of bounds!");
				}
				/*
				if (Val == 0)
					printf("error: digit is 0!\n");
				*/

				if (SAT_TYPE == 2 && Val != 0 && m >= 2)
				{
					printf("warning: you must read files that contain EXACT-2-SAT clauses !!!\n");
					printf("         In the read file there are %d-SAT clauses !\n", m + 1);
					if (m + 1 == 3)
						printf("         You can press 's' in the main menu to switch to 3-SAT.");
					getchar();
					if (F)
						fclose(F);
					return false; // error
				}
				if (SAT_TYPE == 3 && Val != 0 && m >= 3)
				{
					printf("warning: you must read files that contain EXACT-3-SAT clauses !!!\n");
					printf("         In the read file there are %d-SAT clauses !\n", m + 1);
					getchar();
					if (F)
						fclose(F);
					return false; // error
				}
			}

			ClauseIndex ++;
		}

		printf("read %d lines.\n", ClauseIndex);

		ClauseNumber = ClauseIndex;

		if (ClauseNumber > CLAUSE_NUMBER_MAX)
		{
			printf("warning: in the read file, there are too many lines (clauses) !\n");
			getchar();
		}

		for (int m = 0; m < ClauseNumber; m ++)
			for (int n = 0; n < DigitNumber; n ++)
				CNF[m * DigitNumber + n] = CNF[m * DIGIT_NUMBER_MAX + n];
	}
	else
	{
		printf("error reading file !\n");

		ClauseNumber = 0;
	}

	if (F)
		fclose(F);

	return (ClauseNumber > 0);
}

void CNF_CreateRandomlyWithInvertedClauses(const int DigitNumber, const int ClauseNumber, char* CNF) // CNF is changed
{
	for (int i = 0; i < ClauseNumber; i++)
	{
		for (int m = 0; m < DigitNumber; m ++)
			CNF[(i + 0) * DigitNumber + m] = '-';

		if (i + 1 < ClauseNumber)
			for (int m = 0; m < DigitNumber; m ++)
				CNF[(i + 1) * DigitNumber + m] = '-';

		for (int j = 0; j < SAT_TYPE; j ++)
		{
			int Pos;
			while (true)
			{
				Pos = Rand() * DigitNumber / RAND_MAX_PLUS_ONE; // 0...DigitNumber-1
				if (CNF[i * DigitNumber + Pos] == '-')
					break;
			}

			// add original clause
			CNF[(i + 0) * DigitNumber + Pos] = (Rand() < RAND_MAX_SELF_DEF / 2 ? '0' : '1');

			// add additionally the inverted clause
			if (i + 1 < ClauseNumber)
				CNF[(i + 1) * DigitNumber + Pos] = (CNF[(i + 0) * DigitNumber + Pos] == '0' ? '1' : '0'); // invert
		}

		i++;
	}
}

void CNF_CreateRandomlyWithAllPossibleClauses(const int DigitNumber, int& ClauseNumber, char* CNF) // ClauseNumber and CNF is changed
{ // creates a CNF that contains *all* Possible Clauses for a given DigitNumber and SAT type (2/3-SAT)

	if (SAT_TYPE == 2)
	{
		ClauseNumber = 0;

		for (int Digit1 = 1; Digit1 <= DigitNumber - 2; Digit1 ++) // 3 digits per clause
			for (int Digit2 = Digit1 + 1; Digit2 <= DigitNumber - 1; Digit2 ++)
				for (int CIndex = 0; CIndex < 4; CIndex ++)
				{
					for (int i = 0; i < DigitNumber; i ++)
						CNF[ClauseNumber * DigitNumber + i] = '-';

					CNF[ClauseNumber * DigitNumber + (Digit1 - 1)] = (CIndex & 2 ? '1' : '0');
					CNF[ClauseNumber * DigitNumber + (Digit2 - 1)] = (CIndex & 1 ? '1' : '0');

					if (ClauseNumber >= CLAUSE_NUMBER_MAX)
					{
						printf("error in CNF_CreateRandomlyWithAllPossibleClauses(): ClauseNumber too large!\n");
						getchar();
					}
					ClauseNumber ++;
				}

		// randomly swap clauses
		char TempClause[DIGIT_NUMBER_MAX + 1]; // '+ 1' for an eventual null terminator

		for (int i = 0; i < ClauseNumber * 4; i ++)
		{
			// ATTENTION: result of multiplication could exceed int, so use i64!
			int Index1 = (int)((i64)Rand() * (i64)ClauseNumber / (i64)RAND_MAX_PLUS_ONE);
			int Index2 = (int)((i64)Rand() * (i64)ClauseNumber / (i64)RAND_MAX_PLUS_ONE);

			for (int m = 0; m < DigitNumber; m ++)
				TempClause[m] = CNF[Index1 * DigitNumber + m];

			for (int m = 0; m < DigitNumber; m ++)
				CNF[Index1 * DigitNumber + m] = CNF[Index2 * DigitNumber + m];

			for (int m = 0; m < DigitNumber; m ++)
				CNF[Index2 * DigitNumber + m] = TempClause[m];
		}
	}

	if (SAT_TYPE == 3)
	{
		ClauseNumber = 0;

		for (int Digit1 = 1; Digit1 <= DigitNumber - 2; Digit1 ++) // 3 digits per clause
			for (int Digit2 = Digit1 + 1; Digit2 <= DigitNumber - 1; Digit2 ++)
				for (int Digit3 = Digit2 + 1; Digit3 <= DigitNumber; Digit3 ++)
					for (int CIndex = 0; CIndex < 8; CIndex ++)
					{
						for (int i = 0; i < DigitNumber; i ++)
							CNF[ClauseNumber * DigitNumber + i] = '-';

						CNF[ClauseNumber * DigitNumber + (Digit1 - 1)] = (CIndex & 4 ? '1' : '0');
						CNF[ClauseNumber * DigitNumber + (Digit2 - 1)] = (CIndex & 2 ? '1' : '0');
						CNF[ClauseNumber * DigitNumber + (Digit3 - 1)] = (CIndex & 1 ? '1' : '0');

						if (ClauseNumber >= CLAUSE_NUMBER_MAX)
						{
							printf("error in CNF_CreateRandomlyWithAllPossibleClauses(): ClauseNumber too large!\n");
							getchar();
						}
						ClauseNumber ++;
					}

		// randomly swap clauses
		char TempClause[DIGIT_NUMBER_MAX + 1]; // '+ 1' for an eventual null terminator

		for (int i = 0; i < ClauseNumber * 4; i ++)
		{
			// ATTENTION: result of multiplication could exceed int, so use i64!
			int Index1 = (int)((i64)Rand() * (i64)ClauseNumber / (i64)RAND_MAX_PLUS_ONE);
			int Index2 = (int)((i64)Rand() * (i64)ClauseNumber / (i64)RAND_MAX_PLUS_ONE);

			for (int m = 0; m < DigitNumber; m ++)
				TempClause[m] = CNF[Index1 * DigitNumber + m];

			for (int m = 0; m < DigitNumber; m ++)
				CNF[Index1 * DigitNumber + m] = CNF[Index2 * DigitNumber + m];

			for (int m = 0; m < DigitNumber; m ++)
				CNF[Index2 * DigitNumber + m] = TempClause[m];
		}
	}
}

void CNF_CreateSolverCKillerCNF(const int DigitNumber, int& ClauseNumber, char* CNF) // ClauseNumber and CNF is changed
{
	if (SAT_TYPE == 2)
	{
		printf("internal error: Solver C Killer CNF(s) can only be created in 3-SAT mode!\n");
		getchar();
	}

	if (SAT_TYPE == 3) // 'Solver C Killer CNFs' seem to be UNSAT for DigitNumber >= 5
	{
		// NOTE: my thanks go to Mr. Mihai Prunescu, who proposed
		// the following, simplified way of creating the
		// 'Solver C Killer CNFs' - my own method I had told him
		// about was much more (unnecessarily) complicated.

		ClauseNumber = 0;

		for (int P1 = 0; P1 < DigitNumber - 2; P1 ++)
		{
			for (int P2 = P1 + 1; P2 < DigitNumber - 1; P2 ++)
			{
				for (int P3 = P2 + 1; P3 < DigitNumber; P3 ++)
				{
					for (int m = 0; m < DigitNumber; m ++)
						CNF[ClauseNumber * DigitNumber + m] = '-';
					CNF[ClauseNumber * DigitNumber + P1] = '1';
					CNF[ClauseNumber * DigitNumber + P2] = '1';
					CNF[ClauseNumber * DigitNumber + P3] = '1';
					if (ClauseNumber >= CLAUSE_NUMBER_MAX)
					{
						printf("error in CNF_CreateSolverCKillerCNF(): ClauseNumber too large!\n");
						getchar();
					}
					ClauseNumber ++;
					for (int m = 0; m < DigitNumber; m ++)
						CNF[ClauseNumber * DigitNumber + m] = '-';
					CNF[ClauseNumber * DigitNumber + P1] = '0';
					CNF[ClauseNumber * DigitNumber + P2] = '0';
					CNF[ClauseNumber * DigitNumber + P3] = '0';
					if (ClauseNumber >= CLAUSE_NUMBER_MAX)
					{
						printf("error in CNF_CreateSolverCKillerCNF(): ClauseNumber too large!\n");
						getchar();
					}
					ClauseNumber ++;
				}
			}
		}
	}
}

void CNF_CreatePlanted(const int DigitNumber, const int ClauseNumber, char* CNF) // CNF is changed
{ // CNFs created with this procedure are always satisfiable!

	char Solution[DIGIT_NUMBER_MAX];

	// create a random solution
	for (int m = 0; m < DigitNumber; m ++)
		Solution[m] = (Rand() < 16384 ? '0' : '1');

	// build CNF string
	for (int i = 0; i < ClauseNumber; i++)
	{
		for (int m = 0; m < DigitNumber; m ++)
			CNF[i * DigitNumber + m] = '-';

		const int jThatUsesSolution = Rand() * SAT_TYPE / RAND_MAX_PLUS_ONE;

		for (int j = 0; j < SAT_TYPE; j ++)
		{
			int Pos;
			while (true)
			{
				Pos = Rand() * DigitNumber / RAND_MAX_PLUS_ONE; // 0...DigitNumber-1
				if (CNF[i * DigitNumber + Pos] == '-')
					break;
			}
			if (j == jThatUsesSolution)
				CNF[i * DigitNumber + Pos] = Solution[Pos]; // one digit taken from the randomly invented solution...
			else
				CNF[i * DigitNumber + Pos] = (Rand() < 16384 ? '0' : '1'); // ...the other one(s) chosen randomly
		}
	}
}

// *********************** END OF SAT FORMULA CREATION ************************
// ************************** EXPONENTIAL SAT SOLVER **************************

bool CNF_Solve_Exponential(const int DigitNumber, const int ClauseNumber, const char* CNF)
{ // notice that you should NOT USE MORE than 25-30 digits (=DigitNumber) or solving will take very much time!

	// PLEASE NOTE: this solver function may try out
	// *all* possible solutions (if CNF is UNSAT)
	// and thus could need O(2^DigitNumber) time.

	char	Solution[100];
	int		SolutionNumber = 0;
	bool	Solvable = false;

	for (int i = 0; i < DigitNumber; i++)
		Solution[i] = '0';

	while (true)
	{
		// check if *all* clauses of the CNF are satisfied by the current solution:
		for (int ClauseFor = 0; ClauseFor < ClauseNumber; ClauseFor ++)
		{
			bool ClauseSatisfied = false;

			for (int j = 0; j < DigitNumber; j++)
			{
				if (Solution[j] == CNF[ClauseFor * DigitNumber + j])
				{
					ClauseSatisfied = true;
					break;
				}
			}

			if (ClauseSatisfied == false)
				goto NextSolution;
		}

		// if we arrived here, we found a solution
		SolutionNumber ++;

		// show the first 4 solutions
		if (SolutionNumber == 5)
		{
#ifndef MIN_OUTPUT
			printf("[...]\n\n");
#endif
			return true;
		}
		else
		if (SolutionNumber <= 4)
		{
#ifndef MIN_OUTPUT
			printf("(fail-safe) solution:\n");

			for (int j = 0; j < DigitNumber; j ++)
				if (Solution[j] == '0')
					printf("0");
				else
				if (Solution[j] == '1')
					printf("1");

			printf("\n");
#endif
			Solvable = true;
		}

NextSolution: ;
		// switch to next possible solution...
		for (int j = DigitNumber - 1; j >= 0; j--)
		{
			if (Solution[j] == '0')
			{
				Solution[j] = '1';
				goto TestSolution;
			}
			else
			{
				Solution[j] = '0';
			}
		}

		if (Solvable)
		{
#ifndef MIN_OUTPUT
			printf("\n");
#endif
			return true;
		}
		else
		{
			return false;
		}

TestSolution: ;
		// ...and check (in the next while-iteration) if it satisfies the CNF
	}
}

// ********************** END OF EXPONENTIAL SAT SOLVER ***********************
