// (c)2013-2018 by Louis Coder. Implementation of a polynomial 3-SAT solver.

// *****************************
// *** POLYNOMIAL SOLVER >>> ***
// *****************************

// PLEASE NOTE: this code is a variation ('alternative') of the
// original Polynomial Solver vD-1.1 and might therefore contain
// procedures and data structures that were used in the original
// version but not by the variation.
// To allow comparing the code of both versions I kept the original
// code and did merely comment out what must not to be executed in
// this variation.

#define ACTIVE_CLAUSE_TUPLE_NUMBER_MAX			(POSSIBLE_CLAUSE_NUMBER_MAX * POSSIBLE_CLAUSE_NUMBER_MAX)

struct ACTStruct // ActiveClauseTupleStruct; stores one ActiveClauseTuple; used by the Alternative Solver
{
	char			J[DIGIT_NUMBER_MAX + 1]; // plus one char for null terminator
	char			K[DIGIT_NUMBER_MAX + 1];
};

// ******************** GLOBAL POLYNOMIAL SOLVER VARIABLES ********************

// The variables/definitions in this section can be accessed from
// anywhere in each of the Polynomial Solver's procedures.

int					DigitNumberOld = 0;					// used to know when to (re-)initialize PossibleClauses[][], PossibleClauseNumber etc.
int					DigitNumber = 0;					// the clause size of any clause in the CNF, examples: 5 for e.g. "0-00-" or 7 for e.g. "1-0-0--"
int					ClauseNumber = 0;					// number of clauses in the CNF
char*				CNF = NULL;							// pointer to a char string containing the CNF

int					SAT_TYPE_OLD = 0;					// used to detect changes
extern int			SAT_TYPE;							// 2 for 2-SAT or 3 for 3-SAT ('extern' as originally defined in User_Interface.cpp)

int					CombinationCount = 0;				// count of possible 0/1 combinations (see also function 'Solve_Polynomial_Main()')

#define				POSSIBLE_CLAUSE_NUMBER_MAX	(CLAUSE_NUMBER_MAX) // highest PossibleClauseNumber supported by the solver (value gained through tests, see definition of CLAUSE_NUMBER_MAX)

int					PossibleClauseNumber = 0;			// depends on DigitNumber and the choice 2-SAT or 3-SAT
char				PossibleClauses[POSSIBLE_CLAUSE_NUMBER_MAX][DIGIT_NUMBER_MAX + 1]; // see PossibleClauses_Create(); '+ 1' for null terminator

// The following two switches can be turned off to decrease the solving
// time of the Polynomial Solver. PC order shuffling and consistency
// checks can optionally be done to detect errors (i.e. for debugging).

#define				DO_PC_ORDER_SHUFFLING				// enable Possible Clause order shuffling (for testing purposes ONLY - not crucial for solving!)
#define				DO_CONSISTENCY_CHECKS				// enable doing additional checks if data produced by the Polynomial Solver has the expected content

// The 'C[lause]T[able]' data should be used only to do optional verification and
// to produce debug output, because creating and storing the actual Clause Table
// is an O(2^n) task!

#define				CT_DIGIT_NUMBER_MAX			(13)	// 2^13 = 8192 = CT_LINE_NUMBER_MAX; do not create/use/display the ClauseTable if DigitNumber is greater than this!
#define				CT_LINE_NUMBER_MAX			(8192)	// how many lines the ClauseTable can maximal have (2^CT_DIGIT_NUMBER_MAX)
#define				CT_COLUMN_NUMBER_MAX		(384)	// how many columns the ClauseTable can maximal have (approx. CT_DIGIT_NUMBER_MAX * ceiling(CT_DIGIT_NUMBER_MAX / 2) * ceiling(CT_DIGIT_NUMBER_MAX / 4) [plus some safety])

int					CTLineNumber = 0;					// the current count of lines in the ClauseTable
int					CTColumnNumber = 0;					// the current count of columns in the ClauseTable
char				CT[CT_LINE_NUMBER_MAX][CT_COLUMN_NUMBER_MAX][CT_DIGIT_NUMBER_MAX + 1]; // contains clauses; '+ 1' for null terminator
int					CTIndex[CT_LINE_NUMBER_MAX][CT_COLUMN_NUMBER_MAX]; // Possible Clauses index of a clause in ClauseTable

// **************** END OF GLOBAL POLYNOMIAL SOLVER VARIABLES *****************
// ************************** SOLVER HELPER FUNCTIONS *************************

// The following procedures are used by the Polynomial Solver's
// main routine.

bool IsEqual(const char* Clause1, const char* Clause2)
{
	// This function checks if the two char strings are identical.

	for (int m = 0; m < DigitNumber; m ++)
		if ((Clause1[m] != Clause2[m]))
			return false;

	return true;
}

bool IsInConflict(const char* Clause1, const char* Clause2)
{
	// IsInConflict() checks if two clauses are in conflict.
	// Two clauses are in conflict if there's '0' at some position
	// in Clause1 where there's '1' in Clause2 (or vice versa).
	// 
	// For example, it is in conflict:
	// "0-0-" and "1-0-".
	// 
	// Not in conflict are for instance:
	// "0-0-" and "-00-", or
	// "0-0-" and "-0-0", or
	// "0-0-" and "-10-".

	for (int m = 0; m < DigitNumber; m ++)
		if ((Clause1[m] == '0' && Clause2[m] == '1') ||
			(Clause1[m] == '1' && Clause2[m] == '0'))
			return true;

	return false;
}

bool ExistsInCNF(const char* Clause)
{
	// Returns true if Clause appears in the CNF to solve.
	// Makes use of the global variables ClauseNumber,
	// DigitNumber and CNF.

	for (int i = 0; i < ClauseNumber; i ++)
		if (IsEqual(Clause, &CNF[i * DigitNumber]))
			return true;

	return false;
}

bool Does_ih_HaveOneConflict(const char* I, const char* H)
{
	// Returns true if clause I has ONCE some 0 digit where
	// clause H has a 1, or vice versa. If I and H have two
	// or more conflicts, the procedure returns false. If I
	// and H have no conflict, false is returned as well.

	int ConflictPos = -1; // the 'p' position

	for (int m = 0; m < DigitNumber; m ++)
	{
		if ((I[m] == '0' && H[m] == '1') ||
			(I[m] == '1' && H[m] == '0'))
		{
			if (ConflictPos == -1)
				ConflictPos = m;
			else
				return false;
		}
	}

	return true;
}

bool Does_jk_Contain_ih(const char* I, const char* H, const char* J, const char* K)
{
	/*
	Returns true if:
	J+K must cover all 0/1 digits of I, except one
	J+K must cover all 0/1 digits of H, except one

	The resting uncovered I+H 0/1 digits must be in conflict!

	Alternatively true is returned if J+K cover all
	digits of I and H, and there is no conflict at all.

	'C_1+C_2 cover all digits of C_3+C_4' shall mean
	ALL '0' or/and '1' digits of C_3 and C_4 appear
	at the same location in C_1 or/and C_2.
	*/

	int ConflictPos = -1; // the 'p' position

	for (int m = 0; m < DigitNumber; m ++)
	{
		if ((I[m] == '0' && H[m] == '1') ||
			(I[m] == '1' && H[m] == '0'))
		{
			if (ConflictPos == -1)
				ConflictPos = m;
			else
				return false; // more than one conflict
		}
	}

	bool ICoveredHere[DIGIT_NUMBER_MAX];

	for (int m = 0; m < DigitNumber; m ++)
		ICoveredHere[m] = false;

	for (int m = 0; m < DigitNumber; m ++)
		if (J[m] != '-')
			ICoveredHere[m] = true;

	for (int m = 0; m < DigitNumber; m ++)
		if (K[m] != '-')
			ICoveredHere[m] = true;

	for (int m = 0; m < DigitNumber; m ++)
		if (I[m] != '-' && ICoveredHere[m] == false)
			if (m != ConflictPos)
				return false;



	bool HCoveredHere[DIGIT_NUMBER_MAX];

	for (int m = 0; m < DigitNumber; m ++)
		HCoveredHere[m] = false;

	for (int m = 0; m < DigitNumber; m ++)
		if (J[m] != '-')
			HCoveredHere[m] = true;

	for (int m = 0; m < DigitNumber; m ++)
		if (K[m] != '-')
			HCoveredHere[m] = true;

	for (int m = 0; m < DigitNumber; m ++)
		if (H[m] != '-' && HCoveredHere[m] == false)
			if (m != ConflictPos)
				return false;

	return true;
}

// The cache functionality implemented in the following procedures
// is used to speed up the Polynomial Solver by pre-computing data;
// So we do, after a one-time-pre-computing, only need to look up
// one array value instead of entering one or more loops to compute
// the requested data.
// Notice it is required to re-do the pre-computing each time the
// SAT CNF dimensions have changed or even for each new CNF, depen-
// ding on what is pre-computed.

bool IsInConflictCache[POSSIBLE_CLAUSE_NUMBER_MAX][POSSIBLE_CLAUSE_NUMBER_MAX];

void IsInConflict_Cache() // must be called when CNF dimensions have changed
{
	for (int i = 0; i < PossibleClauseNumber; i ++)
		for (int j = 0; j < PossibleClauseNumber; j ++)
			IsInConflictCache[i][j] = IsInConflict(PossibleClauses[i], PossibleClauses[j]);
}

bool IsInConflict(const int PossibleClause1Index, const int PossibleClause2Index)
{
	return IsInConflictCache[PossibleClause1Index][PossibleClause2Index];
}

bool ExistsInCNFCache[POSSIBLE_CLAUSE_NUMBER_MAX];

void ExistsInCNF_Cache() // must be called for every CNF
{
	if (CNF == NULL)
		return; // could be the case if starting program and instantly choosing test mode '7', what will lead to a call of Verify_PossibleClauses_And_Caches_Are_Initialized(), which calls ExistsinCNF_Cache()

	for (int i = 0; i < PossibleClauseNumber; i ++)
		ExistsInCNFCache[i] = ExistsInCNF(PossibleClauses[i]);
}

bool ExistsInCNF(const int PossibleClauseIndex)
{
	return ExistsInCNFCache[PossibleClauseIndex];
}

bool Does_ih_HaveOneConflictCache[POSSIBLE_CLAUSE_NUMBER_MAX][POSSIBLE_CLAUSE_NUMBER_MAX];

void Does_ih_HaveOneConflict_Cache() // must be called when CNF dimensions have changed
{
	for (int i = 0; i < PossibleClauseNumber; i ++)
		for (int h = 0; h < PossibleClauseNumber; h ++)
			Does_ih_HaveOneConflictCache[i][h] = Does_ih_HaveOneConflict(PossibleClauses[i], PossibleClauses[h]);
}

bool Does_ih_HaveOneConflict(const int i, const int h)
{
	return Does_ih_HaveOneConflictCache[i][h];
}

short kInActiveClauseTuple[POSSIBLE_CLAUSE_NUMBER_MAX][POSSIBLE_CLAUSE_NUMBER_MAX + 1]; // plus one for end marker

void kInActiveClauseTuple_Initialize(const bool ActiveClauseTuple[POSSIBLE_CLAUSE_NUMBER_MAX][POSSIBLE_CLAUSE_NUMBER_MAX]) // must be called for every CNF
{
	// Converts the array ActiveClauseTuple into a list of index values.
	// This works like this: kInActiveClauseTuple[j][running number] = k if
	// ActiveClauseTuple[j][k] is true. The end of the list is marked by
	// kInActiveClauseTuple[j][highest running number + 1] == -1.

	for (int j = 0; j < PossibleClauseNumber; j ++)
	{
		int kNumber = 0;
		for (int k = 0; k < PossibleClauseNumber; k ++)
		{
			if (ActiveClauseTuple[j][k])
			{
				kInActiveClauseTuple[j][kNumber] = k;
				kNumber ++;
			}
		}
		kInActiveClauseTuple[j][kNumber] = -1; // end marker
	}
}

void kInActiveClauseTuple_Remove(const int j, bool Removek[POSSIBLE_CLAUSE_NUMBER_MAX])
{
	// Removes all kInActiveClauseTuple[j][x] values for which
	// Removek[x] == true. The intention is to remove multiple elements of
	// kInActiveClauseTuple[j][] with only ONE loop-through to save time.

	int Shift = 0;
	for (int k = 0; k < PossibleClauseNumber; k ++)
	{
		if (Removek[k])
		{
			Removek[k] = false;
			Shift ++;
		}
		else
		{
			if ((kInActiveClauseTuple[j][k - Shift] = kInActiveClauseTuple[j][k]) == -1)
				break;
		}
	}
}

void PossibleClauses_Create()
{
	// Fills the array PossibleClauses[][] and sets PossibleClauseNumber.
	// Those variables are global; the procedure accesses further global
	// variables, namely SAT_TYPE and DigitNumber.

	PossibleClauseNumber = 0;

	if (SAT_TYPE == 2)
	{
		// NOTE: if DigitNumber == 3, the PossibleClauses for 2-SAT are:
		// 00-
		// 01-
		// 10-
		// 11-
		// 0-0
		// 0-1
		// 1-0
		// 1-1
		// -00
		// -01
		// -10
		// -11
		// PossibleClauseNumber = (3 * 4) = 12

		for (int D1 = 0; D1 < DigitNumber - 1; D1 ++) // position of first '0' or '1' digit in Possible Clause
		{
			for (int D2 = D1 + 1; D2 < DigitNumber; D2 ++) // position of second '0' or '1' digit in Possible Clause
			{
				for (int C = 0; C < 4; C ++) // CombinationIndex: 00, 01, 10, 11
				{
					// We arrive here O(DigitNumber^2) times!
					// So there are O(DigitNumber^2) PossibleClauses!

					for (int m = 0; m < DigitNumber; m ++)
						PossibleClauses[PossibleClauseNumber][m] = '-';

					PossibleClauses[PossibleClauseNumber][D1] = (C & 2 ? '1' : '0');
					PossibleClauses[PossibleClauseNumber][D2] = (C & 1 ? '1' : '0');
					PossibleClauses[PossibleClauseNumber][DigitNumber] = '\0';

					if (PossibleClauseNumber >= POSSIBLE_CLAUSE_NUMBER_MAX) // if this happens, try to increase POSSIBLE_CLAUSE_NUMBER_MAX and re-compile
					{
						printf("Warning: PossibleClauseNumber too large in PossibleClauses_Create() !");
						getchar();
					}

					PossibleClauseNumber ++;
				}
			}
		}
	}

	if (SAT_TYPE == 3)
	{
		// NOTE: if DigitNumber == 4, the PossibleClauses for 3-SAT are:
		// 000-
		// 001-
		// 010-
		// 011-
		// 100-
		// 101-
		// 110-
		// 111-
		// 00-0
		// 00-1
		// 01-0
		// 01-1
		// 10-0
		// 10-1
		// 11-0
		// 11-1
		// 0-00
		// 0-01
		// 0-10
		// 0-11
		// 1-00
		// 1-01
		// 1-10
		// 1-11
		// -000
		// -001
		// -010
		// -011
		// -100
		// -101
		// -110
		// -111
		// PossibleClauseNumber = (4 * 8) = 32

		for (int D1 = 0; D1 < DigitNumber - 2; D1 ++) // position of first '0' or '1' digit in Possible Clause
		{
			for (int D2 = D1 + 1; D2 < DigitNumber - 1; D2 ++) // position of second '0' or '1' digit in Possible Clause
			{
				for (int D3 = D2 + 1; D3 < DigitNumber; D3 ++) // position of third '0' or '1' digit in Possible Clause
				{
					for (int C = 0; C < 8; C ++) // 000, 001, 010, 011, 100, 101, 110, 111
					{
						// We arrive here O(DigitNumber^3) times!
						// So there are O(DigitNumber^3) PossibleClauses!

						for (int m = 0; m < DigitNumber; m ++)
							PossibleClauses[PossibleClauseNumber][m] = '-';

						PossibleClauses[PossibleClauseNumber][D1] = (C & 4 ? '1' : '0');
						PossibleClauses[PossibleClauseNumber][D2] = (C & 2 ? '1' : '0');
						PossibleClauses[PossibleClauseNumber][D3] = (C & 1 ? '1' : '0');
						PossibleClauses[PossibleClauseNumber][DigitNumber] = '\0';

						if (PossibleClauseNumber >= POSSIBLE_CLAUSE_NUMBER_MAX) // if this happens, try to increase POSSIBLE_CLAUSE_NUMBER_MAX and re-compile
						{
							printf("Warning: PossibleClauseNumber too large in PossibleClauses_Create() !");
							getchar();
						}

						PossibleClauseNumber ++;
					}
				}
			}
		}
	}
}

// ********************** END OF SOLVER HELPER FUNCTIONS **********************
// ******************************** DEBUG CODE ********************************

// The code in "Polynomial_Solver_Debug_Procedures.h" is not required
// for the polynomial solving process, you could remove the file and
// eventual references to its procedures, and the Polynomial Solver
// would still work.

void ClauseCopy(char* Target, char* Source); // required by Polynomial_Solver_Debug_Procedures.h
void ClauseOverlay(char* Target, char* Source1, char* Source2); // required by Polynomial_Solver_Debug_Procedures.h

#include "Polynomial_Solver_Debug_Procedures.h"

// **************************** END OF DEBUG CODE *****************************
// ************************** SOLVER MAIN PROCEDURE ***************************

void Verify_PossibleClauses_And_Caches_Are_Initialized()
{
	if (DigitNumberOld != DigitNumber ||
		SAT_TYPE_OLD != SAT_TYPE) // at first run or when DigitNumber or/and SAT_TYPE changed
	{
		DigitNumberOld = DigitNumber;
		SAT_TYPE_OLD = SAT_TYPE;

		if (SAT_TYPE == 2)
		{
			CombinationCount = 4; // 00, 01, 10, 11
		}
		if (SAT_TYPE == 3)
		{
			CombinationCount = 8; // 000, 001, 010, 011, 100, 101, 110, 111
		}

		PossibleClauses_Create(); // fill PossibleClauses[][] and set PossibleClauseNumber

#ifndef DO_PC_ORDER_SHUFFLING // if NOT defined!

		// must be done only when PossibleClauses[][] data is to be updated:
		IsInConflict_Cache(); // pre-compute if two arbitrarily chosen Possible Clauses are in conflict
		Does_ih_HaveOneConflict_Cache(); // pre-compute if two Possible Clauses i and h have exactly one conflict

#endif
	}

#ifdef DO_PC_ORDER_SHUFFLING // if IS defined!

	if (Rand2() < 4096) // test some SAT CNFs before doing another shuffling
	{
		// For testing purposes we shuffle the Possible Clause order randomly.
		// Notice that we must keep Possible Clause Types within blocks.
		// Shuffling is NOT REQUIRED for the actual solving, it's just an
		// extra test to verify the Polynomial Solver is not sensitive to
		// Possible Clause Type orders.

		const int RepeatNumber = PossibleClauseNumber / CombinationCount; // can be chosen arbitrarily, take as few as possible (for speed) but enough to do some 'sufficient' shuffling

		for (int Repeat = 0; Repeat < RepeatNumber; Repeat ++)
		{
			int i = Rand2() * PossibleClauseNumber / RAND_MAX_PLUS_ONE;
			int j = Rand2() * PossibleClauseNumber / RAND_MAX_PLUS_ONE;

			i -= (i % CombinationCount); // round down to Possible Clause Type start
			j -= (j % CombinationCount); // round down to Possible Clause Type start

			for (int Block = 0; Block < CombinationCount; Block ++)
			{
				char TempLine[DIGIT_NUMBER_MAX];

				for (int m = 0; m < DigitNumber; m ++)
					TempLine[m] = PossibleClauses[i + Block][m];
				for (int m = 0; m < DigitNumber; m ++)
					PossibleClauses[i + Block][m] = PossibleClauses[j + Block][m];
				for (int m = 0; m < DigitNumber; m ++)
					PossibleClauses[j + Block][m] = TempLine[m];
			}
		}
	}
	else
	if (Rand2() < 4096) // test some SAT CNFs before doing another reversing
	{
		// We reverse the Possible Clauses to check if the Polynomial Solver
		// is sensitive to Possible Clause order. Tests showed that reversing the
		// PC order caused problems in earlier versions of the Polynomial Solver
		// (e.g. in version 'B' from December 2013).

		const int RepeatNumber = PossibleClauseNumber / 2; // must be *exactly* the half!

		for (int Repeat = 0; Repeat < RepeatNumber; Repeat ++)
		{
			int i = Repeat;
			int j = PossibleClauseNumber - 1 - Repeat;

			char TempLine[DIGIT_NUMBER_MAX];

			for (int m = 0; m < DigitNumber; m ++)
				TempLine[m] = PossibleClauses[i][m];
			for (int m = 0; m < DigitNumber; m ++)
				PossibleClauses[i][m] = PossibleClauses[j][m];
			for (int m = 0; m < DigitNumber; m ++)
				PossibleClauses[j][m] = TempLine[m];
		}
	}
	// must be done only when PossibleClauses[][] data is to be updated:
	IsInConflict_Cache(); // pre-compute if two arbitrarily chosen Possible Clauses are in conflict
	Does_ih_HaveOneConflict_Cache(); // pre-compute if two Possible Clauses i and h have exactly one conflict

#endif

	// must be done for each CNF
	ExistsInCNF_Cache(); // pre-compute if a Possible Clause exists in SAT CNF to solve
}

// *** ALTERNATIVE SOLVER >>> ***

bool ExistsInCNFIncluded(const char* Clause)
{
	for (int i = 0; i < ClauseNumber; i ++)
		if (IsEqualOrIncluded(Clause, &CNF[i * DigitNumber]))
			return true;

	return false;
}

void ClauseCopy(char* Target, char* Source)
{
	for (int m = 0; m < DigitNumber; m ++)
		Target[m] = Source[m];
	Target[DigitNumber] = '\0';
}

void ClauseOverlay(char* Target, char* Source1, char* Source2)
{
	// verify
	for (int m = 0; m < DigitNumber; m ++)
		if (Source1[m] != '-' && Source2[m] != '-' && Source1[m] != Source2[m])
		{
			printf("Error in ClauseOverlay(): clauses in conflict!\n");
			getchar();
		}

	// begin
	for (int m = 0; m < DigitNumber; m ++)
		if (Source1[m] != '-')
			Target[m] = Source1[m];
		else
			Target[m] = Source2[m];
	Target[DigitNumber] = '\0';
}

void ThreeOutOf(char* InClause, int& OutNumber, char OutArray[POSSIBLE_CLAUSE_NUMBER_MAX][DIGIT_NUMBER_MAX + 1])
{ // write, to OutArray, all combinations of taking exactly three '0' or/and '1' from InClause, in all possible combinations

	OutNumber = 0;
	for (int c1 = 0; c1 < DigitNumber - 2; c1 ++)
	{
		if (InClause[c1] == '-')
			continue;

		for (int c2 = c1 + 1; c2 < DigitNumber - 1; c2 ++)
		{
			if (InClause[c2] == '-')
				continue;

			for (int c3 = c2 + 1; c3 < DigitNumber - 0; c3 ++)
			{
				if (InClause[c3] == '-')
					continue;

				if (OutNumber >= POSSIBLE_CLAUSE_NUMBER_MAX)
				{
					printf("Too great OutNumber!\n");
					getchar();
				}
				for (int m = 0; m < DigitNumber; m ++)
					OutArray[OutNumber][m] = '-';
				OutArray[OutNumber][c1] = InClause[c1];
				OutArray[OutNumber][c2] = InClause[c2];
				OutArray[OutNumber][c3] = InClause[c3];
				OutArray[OutNumber][DigitNumber] = '\0';
				OutNumber ++;
			}
		}
	}
}

int			ActiveClauseTupleNumber = 0;
ACTStruct	ActiveClauseTuples[ACTIVE_CLAUSE_TUPLE_NUMBER_MAX];

bool IsAnActiveClauseTuple(char* J, char* K)
{
	for (int i = 0; i < ActiveClauseTupleNumber; i ++)
		if (IsEqual(J, ActiveClauseTuples[i].J) &&
			IsEqual(K, ActiveClauseTuples[i].K))
			return true;
	return false;
}

bool Solve_Polynomial_Main() // uses global variables defined at top of this code file!
{
	// treat CNFs whose ClauseTable does not have at least 3 columns specially
	if (SAT_TYPE == 3 && DigitNumber == 3)
	{
		return
			!ExistsInCNF("000") ||
			!ExistsInCNF("001") ||
			!ExistsInCNF("010") ||
			!ExistsInCNF("011") ||
			!ExistsInCNF("100") ||
			!ExistsInCNF("101") ||
			!ExistsInCNF("110") ||
			!ExistsInCNF("111");
	}

	// *** PREPARATION ***

	Verify_PossibleClauses_And_Caches_Are_Initialized();

	/* // for debugging you could now print the Clause Table
	ClauseTable_CreateDynamically();
	ClauseTable_Show();
	getchar();
	*/

	// *** The following large commented-out code block was the
	// non-alternative Polynomial Solver vD-1.0. ***

	// initialize ActiveClauseTuple
	/*
	static bool ActiveClauseTuple[POSSIBLE_CLAUSE_NUMBER_MAX][POSSIBLE_CLAUSE_NUMBER_MAX];
	for (int j = 0; j < PossibleClauseNumber; j ++)
		for (int k = 0; k < PossibleClauseNumber; k ++)
			ActiveClauseTuple[j][k] = !IsInConflict(j, k) && !ExistsInCNF(j) && !ExistsInCNF(k);

	kInActiveClauseTuple_Initialize(ActiveClauseTuple);

	static bool ih_HasNojk[POSSIBLE_CLAUSE_NUMBER_MAX][POSSIBLE_CLAUSE_NUMBER_MAX];
	for (int i = 0; i < PossibleClauseNumber; i ++)
		for (int h = 0; h < PossibleClauseNumber; h ++)
			ih_HasNojk[i][h] = false;

	static bool Removek[POSSIBLE_CLAUSE_NUMBER_MAX];
	for (int i = 0; i < PossibleClauseNumber; i ++)
		Removek[i] = false;

	// *** ENTER SOLVER MAIN LOOPS ***

	bool ChangesExisting = true; // if ActiveClauseTuple[][] content has changed (some kind of 'dirty-flag')

	while (ChangesExisting)
	{
		ChangesExisting = false; // reset

		for (int i = 0; i < PossibleClauseNumber; i ++)
		{
			for (int h = 0; h < PossibleClauseNumber; h ++)
			{
				if (ih_HasNojk[i][h])
					continue;

				ih_HasNojk[i][h] = true;

				if ((!IsInConflict(i, h) || !Does_ih_HaveOneConflict(i, h)) &&
					(i != h)) // i == h IS allowed!!!
					continue;

				for (int j = 0; j < PossibleClauseNumber; j ++)
				{
					if (kInActiveClauseTuple[j][0] < 0)
						continue;

					if (IsInConflict(i, j) ||
						IsInConflict(h, j))
						continue;

					bool	RemovekUsed = false;
					int		kFor = -1;

					while (true)
					{
						const int k = kInActiveClauseTuple[j][++ kFor];
						if (k < 0)
							break;

						if (IsInConflict(i, k) ||
							IsInConflict(h, k)) // IsInConflict(j, k) already 'within' ActiveClauseTuple[j][k], because ActiveClauseTuple[j][k] will be initialized to true only if j not in conflict with k
							continue;

						if ((ActiveClauseTuple[i][j] && ActiveClauseTuple[i][k]) ||
							(ActiveClauseTuple[h][j] && ActiveClauseTuple[h][k]))
						{
							ih_HasNojk[i][h] = false;
							continue;
						}

						// if (!ActiveClauseTuple[j][k]) // already done by kInActiveClauseTuple[j][kFor], which is only true if j/k are in an ActiveClauseTuple!
							// continue;

						if (Does_jk_Contain_ih(
							PossibleClauses[i],
							PossibleClauses[h],
							PossibleClauses[j],
							PossibleClauses[k]))
						{
							/*
							printf("i = %s == false + h = %s == false disabled j = %s, k = %s\n",
								PossibleClauses[i],
								PossibleClauses[h],
								PossibleClauses[j],
								PossibleClauses[k]);
							*/
							/*
							ActiveClauseTuple[j][k] = false;
							Removek[kFor] = true;
							RemovekUsed = true;
						}
					}

					if (RemovekUsed)
					{
						kInActiveClauseTuple_Remove(j, Removek); // sets any Removek[]-element to false if it is true, after processing true value
						ChangesExisting = true;
					}
				}
			}
		}
	}

	// *** DETERMINE SAT OR UNSAT FROM ACTIVECLAUSETUPLE ARRAY ***

	bool SAT = false; // UNSAT - CNF has no solution

	for (int i = 0; i < PossibleClauseNumber; i ++)
	{
		for (int j = 0; j < PossibleClauseNumber; j ++)
		{
			if (ActiveClauseTuple[i][j])
			{
				// printf("still true ActiveClauseTuple: %s,%s\n", PossibleClauses[i], PossibleClauses[j]);
				SAT = true; // SAT - CNF has some solution (we do not know which one)
			}
		}
	}
	*/

	// The alternative polynomial solver:

	if (SAT_TYPE != 3)
	{
		// You should only get here if having altered the source code or
		// having read a non-3-SAT CNF from a file using test mode '2'.
		printf("Warning: This Alternative Polynomial Solver can only solve 3-SAT CNFs.\n");
		printf("This is a limitation of the current implementation, not the algorithm itself.\n");
		printf("You tried to solve a %d-SAT CNF.\n", SAT_TYPE);
		getchar();
		// you might get a lot of further error messages if going on
	}

	ActiveClauseTupleNumber = 0; // reset (don't forget!)

	// collect active tuples
	for (int j = 0; j < PossibleClauseNumber; j ++)
	{
		for (int k = 0; k < PossibleClauseNumber; k ++)
		{
			if (!IsInConflict(j, k) && !ExistsInCNF(j) && !ExistsInCNF(k))
			{
				if (ActiveClauseTupleNumber >= ACTIVE_CLAUSE_TUPLE_NUMBER_MAX)
				{
					printf("ActiveClauseTupleNumber too high!\n");
					getchar();
				}
				ClauseCopy(ActiveClauseTuples[ActiveClauseTupleNumber].J, PossibleClauses[j]);
				ClauseCopy(ActiveClauseTuples[ActiveClauseTupleNumber].K, PossibleClauses[k]);
				ActiveClauseTupleNumber ++;
			}
		}
	}

	// enter solver main loops
	while (true)
	{
		bool ChangesExisting = false;

		// use Rule 1 to eventually remove active tuples (means they get disabled)
		for (int i = 0; i < ActiveClauseTupleNumber; i ++)
		{
			char Overlay[DIGIT_NUMBER_MAX + 1];
			ClauseOverlay(Overlay, ActiveClauseTuples[i].J, ActiveClauseTuples[i].K);
			if (ExistsInCNFIncluded(Overlay))
			{
				ActiveClauseTuples[i].J[0] = '\0';
				ActiveClauseTuples[i].K[0] = '\0';
			}
		}
		// remove disabled tuples (1)
		int Shift = 0;
		for (int i = 0; i < ActiveClauseTupleNumber; i ++) // removes all marked tuples in only one complete iteration!
		{
			if (ActiveClauseTuples[i].J[0] == '\0' ||
				ActiveClauseTuples[i].K[0] == '\0')
			{
				ChangesExisting = true;
				Shift ++;
			}
			else if (Shift > 0)
			{
				ClauseCopy(ActiveClauseTuples[i - Shift].J, ActiveClauseTuples[i].J);
				ClauseCopy(ActiveClauseTuples[i - Shift].K, ActiveClauseTuples[i].K);
			}
		}
		ActiveClauseTupleNumber -= Shift;

		// use Rule 2 to eventually remove active tuples (means they get disabled)
		/*
		disable (j,k) if:
		if ((!ActiveClauseTuple[i][j] || !ActiveClauseTuple[i][k]) &&
            (!ActiveClauseTuple[h][j] || !ActiveClauseTuple[h][k]))
		*/
		for (int t = 0; t < ActiveClauseTupleNumber; t ++)
		{
			for (int Place01Pos = 0; Place01Pos < DigitNumber; Place01Pos ++)
			{
				if (ActiveClauseTuples[t].J[Place01Pos] != '-' ||
					ActiveClauseTuples[t].K[Place01Pos] != '-') // we need to place 0/1 digits at a position within J and K where there's still '-'
					continue;

				char JPlus0[DIGIT_NUMBER_MAX + 1];
				char KPlus0[DIGIT_NUMBER_MAX + 1];

				ClauseCopy(JPlus0, ActiveClauseTuples[t].J);
				ClauseCopy(KPlus0, ActiveClauseTuples[t].K);

				// either (0)
				// JPlus0[Place01Pos] = '0'; // add the 0 digit
				// KPlus0[Place01Pos] = '0';

				char JPlus1[DIGIT_NUMBER_MAX + 1];
				char KPlus1[DIGIT_NUMBER_MAX + 1];

				ClauseCopy(JPlus1, ActiveClauseTuples[t].J);
				ClauseCopy(KPlus1, ActiveClauseTuples[t].K);

				// either (1)
				// JPlus1[Place01Pos] = '1'; // add the 1 digit
				// KPlus1[Place01Pos] = '1';

				char JKOverlay0[DIGIT_NUMBER_MAX + 1];
				ClauseOverlay(JKOverlay0, JPlus0, KPlus0);

				// or (0):
				JKOverlay0[Place01Pos] = '0';

				char JKOverlay1[DIGIT_NUMBER_MAX + 1];
				ClauseOverlay(JKOverlay1, JPlus1, KPlus1);

				// or (1):
				JKOverlay1[Place01Pos] = '1';

				int			IOutOfNumber = 0;
				static char	IOutOf[POSSIBLE_CLAUSE_NUMBER_MAX][DIGIT_NUMBER_MAX + 1];
				int			HOutOfNumber = 0;
				static char	HOutOf[POSSIBLE_CLAUSE_NUMBER_MAX][DIGIT_NUMBER_MAX + 1];

				// the following is only ingenious for 3-SAT clauses;
				// if wanting to solve 2-SAT CNFs, we'd need TwoOutOf()
				ThreeOutOf(JKOverlay0, IOutOfNumber, IOutOf);
				ThreeOutOf(JKOverlay1, HOutOfNumber, HOutOf);

				for (int i = 0; i < IOutOfNumber; i ++)
				{
					for (int h = 0; h < HOutOfNumber; h ++)
					{
						char* I = IOutOf[i];
						char* H = HOutOf[h];

						if ((!IsAnActiveClauseTuple(I, ActiveClauseTuples[t].J) || !IsAnActiveClauseTuple(I, ActiveClauseTuples[t].K)) &&
							(!IsAnActiveClauseTuple(H, ActiveClauseTuples[t].J) || !IsAnActiveClauseTuple(H, ActiveClauseTuples[t].K)))
						{
							ActiveClauseTuples[t].J[0] = '\0';
							ActiveClauseTuples[t].K[0] = '\0';
							goto Nextt; // exit i/h/Place01Pos loops to save work
						}
					}
				}
			}
Nextt:;
		}

		// remove disabled tuples (2)
		Shift = 0;
		for (int i = 0; i < ActiveClauseTupleNumber; i ++) // removes all marked tuples in only one complete iteration!
		{
			if (ActiveClauseTuples[i].J[0] == '\0' ||
				ActiveClauseTuples[i].K[0] == '\0')
			{
				ChangesExisting = true;
				Shift ++;
			}
			else if (Shift > 0)
			{
				ClauseCopy(ActiveClauseTuples[i - Shift].J, ActiveClauseTuples[i].J);
				ClauseCopy(ActiveClauseTuples[i - Shift].K, ActiveClauseTuples[i].K);
			}
		}
		ActiveClauseTupleNumber -= Shift;

		if (!ChangesExisting)
			break;
	}

	// *** DETERMINE SAT OR UNSAT ***

	bool SAT = (ActiveClauseTupleNumber > 0);

	// *** OPTIONALLY DO CONSISTENCY CHECKS ***

#ifdef DO_CONSISTENCY_CHECKS
	// NOTE: the following version of DoConsistencyChecks() has been
	// adapted to the specialties of this ALTERNATIVE solver.
	DoConsistencyChecks(DigitNumber, ClauseNumber,
		ActiveClauseTupleNumber,
		ActiveClauseTuples);
#endif

	return SAT;
}

// *** <<< END OF ALTERNATIVE SOLVER ***

// *********************** END OF SOLVER MAIN PROCEDURE ***********************
// ************************* INTERFACE TO SOLVER USER *************************

bool IsCNFValid(const int DigitNumber, const int ClauseNumber, const char* CNF, const int SAT_TYPE)
{
	// This procedure is called to make sure the Polynomial Solver is used in the way
	// it's meant to be used. This procedure is NOT required for the actual solving!

	switch (SAT_TYPE)
	{
	case 2:

		for (int i = 0; i < ClauseNumber; i ++)
		{
			int UsedDigitNumber = 0;

			for (int j = 0; j < DigitNumber; j ++)
				if (CNF[i * DigitNumber + j] != '-')
					UsedDigitNumber ++;

			if (UsedDigitNumber != 2) // 2
				return false;
		}

		return true;

	case 3:

		for (int i = 0; i < ClauseNumber; i ++)
		{
			int UsedDigitNumber = 0;

			for (int j = 0; j < DigitNumber; j ++)
				if (CNF[i * DigitNumber + j] != '-')
					UsedDigitNumber ++;

			if (UsedDigitNumber != 3) // 3
				return false;
		}

		return true;

	default:

		return false;
	}
}

bool CNF_Solve_Polynomial(const int DigitNumberPassed, const int ClauseNumberPassed, char* CNFPassed, bool& SolvingAbortedFlag) // this is the one and only function to be called by the user of the Polynomial Solver
{ // this is the main procedure of the Polynomial Solver; it returns true if the passed CNF is solvable, false if not

	// Example usage:
	// bool CNFSolvable = CNF_Solve_Polynomial(
	//  4,
	//  8,
	//  "000-"   [NO comma!]
	//  "001-"
	//  "010-"
	//  "011-"
	//  "100-"
	//  "101-"
	//  "110-"
	//  "111-");

	// verify
	if (IsCNFValid(DigitNumberPassed, ClauseNumberPassed, CNFPassed, SAT_TYPE) == false)
	{
		printf("ERROR: something's wrong with the CNF to solve.\n");
		printf("Maybe you tried to solve a 3-SAT CNF in 2-SAT mode\n");
		printf("(or vice versa) or the CNF is neither 2-SAT nor 3-SAT.\n");
		printf("The polynomial solving can NOT be done in this configuration!\n");
		printf("\n");
		printf("Notice that you can press 's' in the main menu to\n");
		printf("switch between the solver's 2-SAT or 3-SAT mode!\n");
		getchar();

		SolvingAbortedFlag = true;
		return false;
	}

	// set the global solver vars
	DigitNumber = DigitNumberPassed;
	ClauseNumber = ClauseNumberPassed;
	CNF = CNFPassed;

	SRand2(RandValue); // to make random number succession of random generator 2 better re-producible

	bool Solvable = Solve_Polynomial_Main(); // the actual Polynomial Solver procedure

	/*
	printf("Used Polynomial Solver to solve:\n");
	CNF_Show(DigitNumber, ClauseNumber, CNF);
	printf("The corresponding ClauseTable is:\n");
	ClauseTable_CreateDynamically(); // initialize Clause Table - used for debugging ONLY as exponential memory/CPU usage
	ClauseTable_Show();
	getchar();
	*/

	// SolvingAbortedFlag = true; // enable to NOT check if Polynomial Solver returned the right result
	
	if (SolvingAbortedFlag)
		printf("WARNING: SolvingAbortedFlag == true (polynomial result is not verified)!\n");

	return Solvable;
}

// ********************* END OF INTERFACE TO SOLVER USER **********************

// ************************************
// *** <<< END OF POLYNOMIAL SOLVER ***
// ************************************
