// (c)2012-2018 by Louis Coder (Matthias Mueller). Alternative implementation 'vD2-1.1' of Polynomial SAT Solver 'vD-1.1'.

// ****************************************************************************
// NOTE: this is a different way to implement my polynomial exact-3-SAT-solving
// algorithm. This code at hand was originally written to test assumptions made
// during the development of the algorithm. As the alternative solver might
// give additional insight into the working mechanisms of the polynomial
// solving algorithm, I decided to publish this alternative solver as well.
// You should have found this alternative version in a zip file downloaded from
// www.louis-coder.com. In this zip file you find the original, non-alternative
// version as well.
// The main difference between original and alternative solver is that the
// alternative version works more closely to the mathematic description of the
// polynomial algorithm, i.e. it uses char strings into which it places 0/1
// digits explicitly and does the overlay operation with char strings. The
// alternative solver was in tests always slower than the original version.
// ****************************************************************************

// Contact: visual_basic_louis@yahoo.com; Visit www.louis-coder.com!
// This program implements a polynomial SAT solver and does also allow to test it.
// The program flow starts in the procedure 'main()' below. The actual solver is
// located in the file 'Polynomial_Solver.h'.

char Version[] = "vD2-1.1"; // released (published) in January 2016 (vD-1.0 in November 2015)

// *********************** MOST IMPORTANT CONFIGURATION ***********************

int		SAT_TYPE =				3;		// 2 for 2-SAT or 3 for 3-SAT (can be changed at runtime by using the 's' command)

#define DN_INIT					8		// the default DigitNumber used when running the program and just hitting 'Return'
#define CN_INIT					40		// the default ClauseNumber used when running the program and just hitting 'Return'

#define TEST_RUN_INDEX_START	0		// in case of an eventual failure, set this index to the one that caused the error (actually used for debugging only)
#define TEST_RUN_NUMBER			100000000
#define TEST_RUN_NUMBER_UNTIL_SIZE_CHANGE \
								10		// after having generated that many CNFs, the CNF dimensions are randomized, if USE_RANDOM_CNF_SIZES is true

bool	USE_RANDOM_CNF_SIZES = true;	// can be changed at runtime by using the 'v' command
int		USE_RANDOM_DN_MIN =		4;		// set a minimal DigitNumber when having USE_RANDOM_CNF_SIZES = true; it must be: 3 <= USE_RANDOM_DN_MIN <= DN_INIT
#define USE_RANDOM_SEED 				// if disabled, random number results and therewith whole test procedures can be re-done when re-starting the solver and 'use' it in the same way as before (using same TestMode, CNF dimensions and so on); if having defined USE_SEED, USE_RANDOM_SEED is ignored
// #define USE_SEED				0   	// force using this random seed (initial value of 'main' random generator (Rand()) *and* 'secondary' random generator (Rand2()))

// #define MIN_OUTPUT  					// enable and re-compile to keep only basic output to speed up solving many (e.g. >1M) CNFs

// NOTE: the alternative solver needs much more memory per data
// unit to store, that's why DIGIT_NUMBER_MAX and CLAUSE_NUMBER_MAX
// is less than in the original implementation.

#define DIGIT_NUMBER_MAX		10		// greatest allowed DigitNumber for reading or creating (random) CNFs
#define CLAUSE_NUMBER_MAX		1030	// greatest allowed ClauseNumber for reading or creating (random) CNFs (2^DIGIT_NUMBER_MAX + some safety is really enough, although this exponential size does not appear during proper use of the solver)

// ******************* END OF MOST IMPORTANT CONFIGURATION ********************
// *************************** LINUX COMPATIBILITY ****************************

#ifdef _WIN32 // should also be defined on Windows x64! (see MSDN)
	#include <tchar.h>			// _TCHAR
	#include <stdio.h>			// fgets
	#include <string.h>			// strlen
	#include <windows.h>		// GetTickCount
	#include <conio.h>			// _kbhit
	#include <math.h>			// pow
#else
	#include <stdio.h>			// fgets
	#include <stdlib.h>			// atoi
	#include <string.h>			// strlen
	#include <math.h>			// pow; needs g++ linker switch '-lm'!
#endif

typedef unsigned long long		u64;	// this var type u64 can hold a very large positive integer number
typedef signed long long		i64;	// this var type i64 can hold a very large positive/negative integer number

// LINUX COMPATIBILITY >>>
// 
#define MAX(a, b)				((a) > (b) ? (a) : (b)) // usable in both Windows and Linux compilations

#ifdef _WIN32
	#define GETS(s)				gets_s(s, sizeof(s) - 1)
	#define SPRINTF1(d, f, a1)	sprintf_s(d, sizeof(d) - 1, f, a1)
	#define SPRINTF2(d, f, a1, a2) \
								sprintf_s(d, sizeof(d) - 1, f, a1, a2)
#else
	// NOTE: to make the original Windows code compile on Linux
	// (did compile in tests on Ubuntu 14.04 LTS), we 'imitate'
	// some Windows-exclusive procedures.
	// 
	// IMPORTANT: when using g++ to compile, you need to use the
	// linker switch '-lm', or you get an error message that pow()
	// would not be defined (although it is within math.h).
	// 
	// The present code did compile on Ubuntu 14.04 LTS by entering
	// the following command in a terminal ("Ctrl-Alt-T"):
	// g++ User_Interface.cpp -Xlinker -lm -o psat
	// 
	// Run compiled binary by typing into terminal:
	// cd [directory the compiled "psat" binary is located in]
	// ./psat
	// 
	// If you have problems please google or/and ask question(s) in
	// any Ubuntu forum(s).

	#define DWORD				int
	#define GETS(s)				if (fgets(s, sizeof(s) - 1, stdin)) {} // we need some useless return value check to prevent G++ (Linux) from outputting a warning
	#define SPRINTF1(d, f, a1)	sprintf(d, f, a1)
	#define SPRINTF2(d, f, a1, a2) \
								sprintf(d, f, a1, a2)
	#define ULLONG_MAX			0xffffffffffffffff

	// http://stackoverflow.com/questions/2958291/equivalent-to-gettickcount-on-linux (2015-08-10, altered):
	#include <time.h>			// timespec
	int GetTickCount()
	{
		struct timespec ts;
		if (clock_gettime(CLOCK_MONOTONIC, &ts) != 0)
		{
			return 0;
		}
		return (int)ts.tv_sec * 1000; // we do, here in Linux, only measure in seconds instead of milliseconds; doesn't matter, seconds are precise enough
	}

	#include <unistd.h>
	void Sleep(int ms)
	{
		sleep((ms + 999) / 1000); // sleep takes seconds, Sleep milliseconds!
	}

	// http://cboard.cprogramming.com/c-programming/63166-kbhit-linux.html (2015-08-10, altered):
	#include <stdio.h>
	#include <termios.h>
	#include <unistd.h>
	#include <fcntl.h>

	int _kbhit(void)
	{
		struct termios oldt, newt;
		int ch;
		int oldf;

		tcgetattr(STDIN_FILENO, &oldt);
		newt = oldt;
		newt.c_lflag &= ~(ICANON | ECHO);
		tcsetattr(STDIN_FILENO, TCSANOW, &newt);
		oldf = fcntl(STDIN_FILENO, F_GETFL, 0);
		fcntl(STDIN_FILENO, F_SETFL, oldf | O_NONBLOCK);

		ch = getchar();

		tcsetattr(STDIN_FILENO, TCSANOW, &oldt);
		fcntl(STDIN_FILENO, F_SETFL, oldf);

		if (ch != EOF)
		{
			// ungetc(ch, stdin); // disabled by me (Matthias/Louis)
			return 1;
		}

		return 0;
	}
#endif

// ************************ END OF LINUX COMPATIBILITY ************************
// ************************** USER CONSOLE INTERFACE **************************

#define TEST_MODE_OTHER					-1
#define TEST_MODE_EXIT					0
#define TEST_MODE_RANDOM_CNF			1
#define TEST_MODE_FILES					2
#define TEST_MODE_RANDOM_INVERTED_CNF	3
#define TEST_MODE_COMPLETE_CNF			4
#define TEST_MODE_PIGEON_2				5
#define TEST_MODE_PIGEON_3				6
#define TEST_MODE_ALL_COMPLETE			7
#define TEST_MODE_SOLVER_C_KILLER_CNF	8
#define TEST_MODE_PLANTED_CNF			9

#include "CNF_Creation.h"
#include "Polynomial_Solver.h" // the C++ compiler will insert here the file containing the actual Polynomial Solver!

// FOR TEST_MODE_ALL_COMPLETE >>>
int TestMode_7_PossibleClausesOnOffCombinationIndex;
// <<< END OF FOR TEST_MODE_ALL_COMPLETE

#ifdef _WIN32
int _tmain(int argc, _TCHAR* argv[]) // the program starts here
#else
int main(int argc, char* argv[]) // the program starts here
#endif
{
	// The following array saves the CNF that is to be solved.
	// In 'CNF', clauses appear as coherent string, one clause after another.
	// CNF[ClauseIndex * DigitNumber + DigitIndex] is the DigitIndex-th char
	// in the ClauseIndex-th clause. The content of CNF is passed to the
	// exponential- and to the polynomial solver.

	static char	CNF[CLAUSE_NUMBER_MAX * DIGIT_NUMBER_MAX]; // possible digits: '0', '1', '-'
	
	// Use either same random seed for same re-producible test runs, or
	// (practically) a random seed (via GetTickCount()) for generating
	// always new CNFs in each program instance.

#ifdef USE_SEED
	int RandomSeed = (USE_SEED);
#else
#ifdef USE_RANDOM_SEED
	int RandomSeed = (GetTickCount() % 65536);
#else
	int RandomSeed = (0);
#endif
#endif

	while (true) // the endless-loop that processes user input
	{
		int		TestMode;
		int		TestRunNumber;
		int		TestRunStartIndex;

		int		DN; // digit number of CNF to solve, desired by user
		int		CN; // clause number of CNF to solve, desired by user

		char	FilePathFormat[1024]; // used for test mode '2' (read CNF from file) only

		bool	ShowCNFFlag;

		char	S[1024]; // stores the user input from gets()

		while (true)
		{
			// SET STANDARD TEST (just hit Return to run) >>>:
			
			TestMode = TEST_MODE_RANDOM_CNF;
			TestRunNumber = TEST_RUN_NUMBER;
			TestRunStartIndex = 1;

			DN = DN_INIT;
			CN = CN_INIT;

			ShowCNFFlag = false;

			// <<< END OF SETTING STANDARD TEST

			printf("**********************************************************************\n");
			printf("            POLYNOMIAL EXACT-2/3-SAT SOLVER DEMO %s\n", Version);
			printf("                    * From: www.louis-coder.com\n");
			printf(" * This program can do a mass-test if results of the polynomial solver\n");
			printf("         match the results of a fail-safe, exponential solver\n");
			printf("              * Please read documentation for details.\n");
			printf("**********************************************************************\n");
			printf("\n");
			printf("THIS IS THE ALTERNATIVE SOLVER %s THAT IMPLEMENTS THE POLYNOMIAL\n", Version);
			printf("EXACT-3-SAT-SOLVING ALGORITHM BY MATTHIAS MUELLER IN AN ALTERNATIVE WAY.\n");
			printf("Please view source code for more information.\n");
			printf("\n");
			/* // Alternative Solver cannot handle 2-SAT CNFs
			if (SAT_TYPE == 2)
			{
				printf("TWO-SAT: Works as 2-SAT solver.\n");
				printf("Press 's' to switch to 3-SAT.\n");
			}
			*/
			if (SAT_TYPE == 3)
			{
				printf("THREE-SAT: Works as 3-SAT solver.\n");
				printf("Alternative Solver %s can solve 3-SAT CNFs only.\n", Version);
				printf("(This is no restriction of the polynomial algorithm but of the implementation.)\n");
				// printf("Press 's' to switch to 2-SAT.\n"); // Alternative Solver cannot handle 2-SAT CNFs
			}
			printf("\n");
			printf("Please select testing mode:\n");
			printf("\n");
			printf("1: test random CNFs\n");
			printf("2: test CNF from file\n");
			printf("3: test random CNFs with each clause additionally negated\n");
			printf("4: test 'complete CNFs' (contain all possible clauses, shuffled)\n");
			printf("5: test 'pigeon hole problem' n=2, converted to pure 3-SAT\n");
			// printf("6: test 'pigeon hole problem' n=3, converted to pure 3-SAT\n"); // CNF dimensions too large for this solver
			printf("7: test all build-able CNFs of a given size\n");
			printf("8: test 'Solver C Killer CNFs' that made my previous solver 'C' fail\n");
			printf("9: test 'planted SAT' CNFs\n");
			printf("\n");
			if (USE_RANDOM_CNF_SIZES)
				printf("v: toggle USE_RANDOM_CNF_SIZES (currently on)\n");
			else
				printf("v: toggle USE_RANDOM_CNF_SIZES (currently off)\n");
			printf("e: exit\n");
			printf("\n");
			printf("Just hit Enter to run standard test\n");
			printf("(random CNFs, DigitNumber=%d, ClauseNumber = %d, %d test runs)\n", DN, CN, TestRunNumber);

			GETS(S);

			switch (S[0])
			{
			case 'e':
				TestMode = TEST_MODE_EXIT;
				break;

				/*
			case 's':
				if (SAT_TYPE == 2)
				{
					SAT_TYPE = 3;
					printf("Switched to THREE-SAT.\n");
					Sleep(2000);
					printf("\n");
					continue;
				}
				if (SAT_TYPE == 3)
				{
					SAT_TYPE = 2;
					printf("Switched to TWO-SAT.\n");
					Sleep(2000);
					printf("\n");
					continue;
				}
				continue; // should not happen
				*/

			case 'v':
				if (USE_RANDOM_CNF_SIZES)
				{
					USE_RANDOM_CNF_SIZES = false;
					printf("USE_RANDOM_CNF_SIZES switched off.\n");
					Sleep(2000);
					printf("\n");
					continue;
				}
				else
				{
					USE_RANDOM_CNF_SIZES = true;
					printf("USE_RANDOM_CNF_SIZES switched on.\n");
					Sleep(2000);
					printf("\n");
					continue;
				}
				continue; // should not happen

			case '1':
				printf("**************************************************************\n");
				printf("* The solver will generate random CNFs which are solved by   *\n");
				printf("* first the exponential 'brute-force' algorithm and then by  *\n");
				printf("* the polynomial one. The results are checked for equality.  *\n");
				printf("*                                                            *\n");
				printf("* Tip: do not use a DigitNumber much greater than 10 as the  *\n");
				printf("* polynomial solving has a complexity of > O(DigitNumber^18) *\n"); // the *Alternative* Polynomial Solver is significantly slower than the actual one introduced in the paper, due to missing measurements we just say the Alternative (!) Polynomial Solver's complexity is "greater than O(DigitNumber^18)"
				printf("* and is therefore slow on common PCs.                       *\n");
				printf("*                                                            *\n");
				printf("* Notice: the solver tells only *if* the CNF is solvable,    *\n");
				printf("* it does not find any explicit solution(s).                 *\n");
				printf("**************************************************************\n");
				printf("\n");
				if (SAT_TYPE == 2)
				{
					printf("DigitNumber       (e.g. 12): ");
					GETS(S);
					DN = atoi(S);
					printf("ClauseNumber      (e.g. 24): ");
				}
				if (SAT_TYPE == 3)
				{
					printf("DigitNumber        (e.g. 6): ");
					GETS(S);
					DN = atoi(S);
					printf("ClauseNumber      (e.g. 30): ");
				}
				GETS(S);
				CN = atoi(S);
				printf("TestRunNumber (e.g. 100000): ");
				GETS(S);
				TestRunNumber = atoi(S);
				TestRunStartIndex = 1;
				printf("Show CNFs? y/n     (e.g. n): ");
				GETS(S);
				ShowCNFFlag = (S[0] == 'y');

				TestMode = TEST_MODE_RANDOM_CNF;
				break;

			case '2':
				if (USE_RANDOM_CNF_SIZES)
				{
					USE_RANDOM_CNF_SIZES = false;
					printf("USE_RANDOM_CNF_SIZES switched off.\n");
					Sleep(2000);
					printf("\n");
				}
				printf("****************************************************************************\n");
				printf("* The solver is able to read CNFs from files that have the DIMACS format.  *\n");
				printf("* See 'http://www.cs.ubc.ca/~hoos/SATLIB/Benchmarks/SAT/satformat.ps'.     *\n");
				printf("*                                                                          *\n");
				printf("* You can enter the relative or absolute path to the file that contains    *\n");
				printf("* the SAT CNF to be solved.                                                *\n");
				printf("*                                                                          *\n");
				printf("* Please notice that the solver reads as many clauses as there are in      *\n");
				printf("* the file, meta data or/and headers are ignored.                          *\n");
				printf("*                                                                          *\n");
				printf("* Also notice that the solver tells only *if* the CNF is solvable,         *\n");
				printf("* it does not find any explicit solution(s).                               *\n");
				printf("****************************************************************************\n");
				printf("\n");
				printf("Path            : ");
				GETS(FilePathFormat);
				for (int m = 0; m < 1024; m ++)
				{
					if (FilePathFormat[m] == '\r' ||
						FilePathFormat[m] == '\n')
					{
						FilePathFormat[m] = '\0'; // there were some problems under Ubuntu, file to read not found as FilePathFormat contained tailing \r or/and \n
						break;
					}
				}
				DN = 10; // dummy value; will be changed when reading file
				CN = 10; // dummy value; will be changed when reading file
				TestRunNumber = 1;
				TestRunStartIndex = 1;

				TestMode = TEST_MODE_FILES;
				break;

			case '3':
				printf("**************************************************************\n");
				printf("* The solver will generate random CNFs which are solved by   *\n");
				printf("* first the exponential 'brute-force' algorithm and then by  *\n");
				printf("* the polynomial one. The results are checked for equality.  *\n");
				printf("*                                                            *\n");
				printf("* The CNFs are generated in such a way so that each clause's *\n");
				printf("* literals appear also negated in a corresponding further    *\n");
				printf("* clause.                                                    *\n");
				printf("*                                                            *\n");
				printf("* I realized that solving this kind of CNFs lead to more     *\n");
				printf("* errors in early versions of the polynomial solver, so I    *\n");
				printf("* regard this kind of CNFs as 'relatively hard to solve'.    *\n");
				printf("*                                                            *\n");
				printf("* Tip: do not use a DigitNumber much greater than 10 as the  *\n");
				printf("* polynomial solving has a complexity of > O(DigitNumber^18) *\n"); // the *Alternative* Polynomial Solver is significantly slower than the actual one introduced in the paper, due to missing measurements we just say the Alternative (!) Polynomial Solver's complexity is "greater than O(DigitNumber^18)"
				printf("* and is therefore slow on common PCs.                       *\n");
				printf("*                                                            *\n");
				printf("* Notice: the solver tells only *if* the CNF is solvable,    *\n");
				printf("* it does not find any explicit solution(s).                 *\n");
				printf("**************************************************************\n");
				printf("\n");
				if (SAT_TYPE == 2)
				{
					printf("DigitNumber       (e.g. 12): ");
					GETS(S);
					DN = atoi(S);
					printf("ClauseNumber      (e.g. 24): ");
				}
				if (SAT_TYPE == 3)
				{
					printf("DigitNumber        (e.g. 6): ");
					GETS(S);
					DN = atoi(S);
					printf("ClauseNumber      (e.g. 30): ");
				}
				GETS(S);
				CN = atoi(S);
				printf("TestRunNumber (e.g. 100000): ");
				GETS(S);
				TestRunNumber = atoi(S);
				TestRunStartIndex = 1;
				printf("Show CNFs? y/n     (e.g. n): ");
				GETS(S);
				ShowCNFFlag = (S[0] == 'y');

				TestMode = TEST_MODE_RANDOM_INVERTED_CNF;
				break;

			case '4':
				if (USE_RANDOM_CNF_SIZES)
				{
					USE_RANDOM_CNF_SIZES = false;
					printf("USE_RANDOM_CNF_SIZES switched off.\n");
					Sleep(2000);
					printf("\n");
				}
				printf("**************************************************************\n");
				printf("* The solver will generate random CNFs which are solved by   *\n");
				printf("* first the exponential 'brute-force' algorithm and then by  *\n");
				printf("* the polynomial one. The results are checked for equality.  *\n");
				printf("*                                                            *\n");
				printf("* The created CNFs will consist of all possible clauses a    *\n");
				printf("* CNF with the entered dimensions can have. Those CNFs are   *\n");
				printf("* always unsatisfiable, that means they do not have any      *\n");
				printf("* solution. The CNFs' clauses are permuted randomly.         *\n");
				printf("*                                                            *\n");
				printf("* Tip: do not use a DigitNumber much greater than 10 as the  *\n");
				printf("* polynomial solving has a complexity of > O(DigitNumber^18) *\n"); // the *Alternative* Polynomial Solver is significantly slower than the actual one introduced in the paper, due to missing measurements we just say the Alternative (!) Polynomial Solver's complexity is "greater than O(DigitNumber^18)"
				printf("* and is therefore slow on common PCs.                       *\n");
				printf("*                                                            *\n");
				printf("* Notice: the solver tells only *if* the CNF is solvable,    *\n");
				printf("* it does not find any explicit solution(s).                 *\n");
				printf("**************************************************************\n");
				printf("\n");
				printf("DigitNumber       (e.g. 10): ");
				GETS(S);
				DN = atoi(S);
				CN = 1; // will be set later
				printf("TestRunNumber (e.g. 100000): ");
				GETS(S);
				TestRunNumber = atoi(S);
				TestRunStartIndex = 1;
				printf("Show CNFs? y/n     (e.g. n): ");
				GETS(S);
				ShowCNFFlag = (S[0] == 'y');

				if (DN > DIGIT_NUMBER_MAX) // verify
					DN = 0; // will lead to an error message

				TestMode = TEST_MODE_COMPLETE_CNF;
				break;

			case '5':
				if (SAT_TYPE == 2)
				{
					SAT_TYPE = 3;
					printf("Switched to THREE-SAT.\n");
					Sleep(2000);
					printf("\n");
				}
				if (USE_RANDOM_CNF_SIZES)
				{
					USE_RANDOM_CNF_SIZES = false;
					printf("USE_RANDOM_CNF_SIZES switched off.\n");
					Sleep(2000);
					printf("\n");
				}

				if (SAT_TYPE == 3)
				{
					printf("*******************************************************************************\n");
					printf("* The pigeon hole problem is a kind of CNF that makes resolution-based        *\n");
					printf("* solvers need exponentially much time for solving.                           *\n");
					printf("*                                                                             *\n");
					printf("* As (among other things) the polynomial solver you are currently using is    *\n");
					printf("* not a resolution-based solver, it can detect the pigeon hole instances as   *\n");
					printf("* unsatisfiable within polynomial time and space.                             *\n");
					printf("*                                                                             *\n");
					printf("* Unfortunately this polynomial solver cannot solve large pigeon hole         *\n");
					printf("* problems within a reasonable time on common PCs as the solver generally     *\n");
					printf("* needs > O(n^18) time. Therefore in the following test a rather small        *\n"); // the *Alternative* Polynomial Solver is significantly slower than the actual one introduced in the paper, due to missing measurements we just say the Alternative (!) Polynomial Solver's complexity is "greater than O(DigitNumber^18)"
					printf("* pigeon hole problem instance is solved for demonstration purposes.          *\n");
					printf("*******************************************************************************\n");
					printf("\n");
					printf("Press Return to begin test...\n");
					printf("\n");

					getchar();

					TestMode = TEST_MODE_PIGEON_2;
					TestRunNumber = 1;
					TestRunStartIndex = 1;

					DN = 7;
					CN = 12;

					ShowCNFFlag = true;
				}

				break;

				/* // disabled as the Alternative Polynomial Solver is too slow and needs too much memory for solving a CNF with DigitNumber = 13
			case '6':
				if (SAT_TYPE == 2)
				{
					SAT_TYPE = 3;
					printf("Switched to THREE-SAT.\n");
					Sleep(2000);
					printf("\n");
				}
				if (USE_RANDOM_CNF_SIZES)
				{
					USE_RANDOM_CNF_SIZES = false;
					printf("USE_RANDOM_CNF_SIZES switched off.\n");
					Sleep(2000);
					printf("\n");
				}

				if (SAT_TYPE == 3)
				{
					printf("*******************************************************************************\n");
					printf("* The pigeon hole problem is a kind of CNF that makes resolution-based        *\n");
					printf("* solvers need exponentially much time for solving.                           *\n");
					printf("*                                                                             *\n");
					printf("* As (among other things) the polynomial solver you are currently using is    *\n");
					printf("* not a resolution-based solver, it can detect the pigeon hole instances as   *\n");
					printf("* unsatisfiable within polynomial time and space.                             *\n");
					printf("*                                                                             *\n");
					printf("* Unfortunately this polynomial solver cannot solve large pigeon hole         *\n");
					printf("* problems within a reasonable time on common PCs as the solver generally     *\n");
					printf("* needs O(n^18) time. Therefore in the following test a rather small          *\n");
					printf("* pigeon hole problem instance is solved for demonstration purposes.          *\n");
					printf("*******************************************************************************\n");
					printf("\n");
					printf("Press Return to begin test...\n");
					printf("\n");

					getchar();

					TestMode = TEST_MODE_PIGEON_3;
					TestRunNumber = 1;
					TestRunStartIndex = 1;

					DN = 13;
					CN = 26;

					ShowCNFFlag = true;
				}

				break;
				*/

			case '7':
				if (USE_RANDOM_CNF_SIZES)
				{
					USE_RANDOM_CNF_SIZES = false;
					printf("USE_RANDOM_CNF_SIZES switched off.\n");
					Sleep(2000);
					printf("\n");
				}
				do
				{
					printf("**************************************************************\n");
					printf("* The solver will generate and solve all CNFs build-able     *\n");
					printf("* having the DigitNumber you entered. DigitNumber is the     *\n");
					printf("* literal index range of the literals within the CNFs.       *\n");
					printf("*                                                            *\n");
					printf("* Notice that it can take a long time to generate and solve  *\n");
					printf("* all possible CNFs if the DigitNumber is large.             *\n");
					printf("*                                                            *\n");
					printf("* Because of this, there are the following limits:           *\n");
					printf("*                                                            *\n");
					// printf("* 2-SAT: DigitNumber = 3 or 4,                               *\n"); // 2-SAT not supported by the alternative solver
					printf("* 3-SAT: DigitNumber = 3.                                    *\n");
					printf("**************************************************************\n");
					printf("\n");
					printf("DigitNumber        (e.g. 3): ");
					GETS(S);
					printf("\n");

					DN = atoi(S);
					CN = 1; // dummy value; will be set later
				}
				while (!(
					(SAT_TYPE == 2 && DN >= 3 && DN <= 4) ||
					(SAT_TYPE == 3 && DN >= 3 && DN <= 3)));

				TestMode = TEST_MODE_ALL_COMPLETE;

				if (SAT_TYPE == 2 && DN == 3)
					TestRunNumber = (int)pow(2.0, 4.0 * 3.0); // PossibleClauseNumber = 4 0/1 combinations * 3 ClauseTableColumns...
				if (SAT_TYPE == 2 && DN == 4)
					TestRunNumber = (int)pow(2.0, 4.0 * 6.0); // ...and there are 2^PossibleClauseNumber on/off combinations of PossibleClauses
				if (SAT_TYPE == 3 && DN == 3)
					TestRunNumber = (int)pow(2.0, 8.0 * 1.0); // in 3-SAT CNFs, we have 8 instead of 4 0/1 combinations!

				TestRunStartIndex = 1;

				ShowCNFFlag = true;

				TestMode_7_PossibleClausesOnOffCombinationIndex = 1;
				break;

			case '8':
				if (SAT_TYPE == 2)
				{
					SAT_TYPE = 3;
					printf("Switched to THREE-SAT.\n");
					Sleep(2000);
					printf("\n");
				}
				if (USE_RANDOM_CNF_SIZES)
				{
					USE_RANDOM_CNF_SIZES = false;
					printf("USE_RANDOM_CNF_SIZES switched off.\n");
					Sleep(2000);
					printf("\n");

				}
				printf("***************************************************************\n");
				printf("* Make this solver '%s' solve the such called            *\n""""", Version);
				printf("* 'Solver C Killer CNFs', which made my previous polynomial   *\n");
				printf("* solver version 'C-1.0' return wrong results.                *\n");
				printf("*                                                             *\n");
				printf("* The special property of the Killer-CNFs is that the solver  *\n");
				printf("* cannot save 'intermediate-data' within merely three 0/1     *\n");
				printf("* digits, like fatally done in solver version 'C-1.0'.        *\n");
				printf("***************************************************************\n");
				printf("\n");
				printf("DigitNumber        (e.g. 8): ");
				GETS(S);
				DN = atoi(S);
				printf("TestRunNumber (e.g. 100000): ");
				GETS(S);
				// 
				// NOTE: it *does* make sense to do more than one test run if you
				// have 'DO_PC_ORDER_SHUFFLING' in Polynomial_Solver.h defined,
				// because the Polynomial Solver then involves a random number
				// generator and could theoretically fail, depending on the
				// current Possible Clauses order.
				// 
				TestRunNumber = atoi(S);
				TestRunStartIndex = 1;
				printf("Show CNFs? y/n     (e.g. n): ");
				GETS(S);
				ShowCNFFlag = (S[0] == 'y');

				TestMode = TEST_MODE_SOLVER_C_KILLER_CNF;
				break;

			case '9':
				printf("********************************************************************\n");
				printf("* The program will, for each CNF, first invent a random solution   *\n");
				printf("* (which is in literature mostly called 'model').                  *\n");
				printf("*                                                                  *\n");
				printf("* Then the actual CNF is generated in such a way that in each      *\n");
				printf("* clause there is one literal that has the negation state of the   *\n");
				printf("* related digit from the random solution. The resting 2 literals   *\n");
				printf("* (for 3-SAT) are randomly negated or not. The indices of the      *\n");
				printf("* literals (i.e. positions of 0/1 digits within the clauses) are   *\n");
				printf("* selected randomly, too.                                          *\n");
				printf("*                                                                  *\n");
				printf("* Like this, always-solvable CNFs are generated.                   *\n");
				printf("********************************************************************\n");
				printf("\n");
				if (SAT_TYPE == 2)
				{
					printf("DigitNumber       (e.g. 12): ");
					GETS(S);
					DN = atoi(S);
					printf("ClauseNumber      (e.g. 24): ");
				}
				if (SAT_TYPE == 3)
				{
					printf("DigitNumber        (e.g. 6): ");
					GETS(S);
					DN = atoi(S);
					printf("ClauseNumber      (e.g. 30): ");
				}
				GETS(S);
				CN = atoi(S);
				printf("TestRunNumber (e.g. 100000): ");
				GETS(S);
				TestRunNumber = atoi(S);
				TestRunStartIndex = 1;
				printf("Show CNFs? y/n     (e.g. n): ");
				GETS(S);
				ShowCNFFlag = (S[0] == 'y');

				TestMode = TEST_MODE_PLANTED_CNF;
				break;

			case '\0': // user just hit Return; run standard test
				break;
			case '\r': // user just hit Return; run standard test (Linux)
				break;
			case '\n': // user just hit Return; run standard test (Linux)
				break;

			default: // unknown user input
				continue; // show info text again
			}

			if (TestMode == TEST_MODE_EXIT || (
				TestMode >= TEST_MODE_RANDOM_CNF && TestMode <= TEST_MODE_PLANTED_CNF &&
				DN >= 3 && DN <= DIGIT_NUMBER_MAX &&
				CN >= 1 && CN <= CLAUSE_NUMBER_MAX &&
				TestRunNumber >= 1 &&
				TestRunNumber <= 100000000 &&
				TestRunStartIndex >= 1 &&
				TestRunStartIndex <= 100000000))
			{
				printf("\n");
				break;
			}
			else
			{
				printf("\n");
				printf("ERROR: Wrong input !\n");
				printf("       Probably you tried to exceed the CNF size limits.\n");
				printf("       Try again with other parameters.\n");
				printf("       DigitNumberMax=%d, ClauseNumberMax=%d,\n", DIGIT_NUMBER_MAX, CLAUSE_NUMBER_MAX);
				printf("       TestRunNumberMax=%d.\n", 100000000);
				Sleep(2000);
				printf("\n");
			}
		}

		if (TestMode == TEST_MODE_EXIT)
			break;

		// for info-statistics:
		int		SuccessNumber = 0;
		int		ErrorNumber = 0;
		int		UnknownNumber = 0;

		int		SolvableNumber = 0;
		int		UNSATNumber = 0;

		printf("\n");

		// initialize main random generator used to 'invent' CNFs
		SRand(RandomSeed); // same SRand() (and SRand2()) parameter means same value succession of Rand() every time you run the program

		// also initialize the second random generator to get reproducible random number sets
		SRand2(RandomSeed);

		// enter loop that reads/invents SAT CNFs and that checks those CNFs' solvability
		// once with an exponential solver and then with the polynomial one - the result
		// pairs are each checked for equality (they must always be equal)

		int		DN_Original = DN; // back up for the case we use USE_RANDOM_CNF_SIZES
		int		CN_Original = CN; // back up for the case we use USE_RANDOM_CNF_SIZES

		for (int TestRunIndex = TestRunStartIndex; TestRunIndex < TestRunStartIndex + TestRunNumber; TestRunIndex += 1)
		{
			if (TestRunIndex >= TEST_RUN_INDEX_START) // for debugging
			{
				printf("\n*********************************************************** TestRun #%d\n", TestRunIndex);
				if (RandomSeed != 0)
					printf("RandomSeed=%d\n", RandomSeed); // if an error appears, you must change the code and pass this RandomSeed to SRand() and SRand2() to re-produce the error (can be done via #define USE_SEED ..., see top of this file)
			}

			// vary CNF dimensions if enabled (use 'v' command in console to toggle enabled state)
			if (USE_RANDOM_CNF_SIZES &&
				((TestRunIndex % TEST_RUN_NUMBER_UNTIL_SIZE_CHANGE) == 0 || (TestRunIndex == 1))) // change CNF dimensions only every 25th CNF to avoid permanent, CPU-consuming re-calling PossibleClauses_Create() in Solve_Polynomial_Main()
			{
				// The following varies the CNF dimensions every TEST_RUN_NUMBER_UNTIL_SIZE_CHANGE-th
				// CNF in a defined range:
				// DN = USE_RANDOM_DN_MIN...DN_Original and CN is calculated in a way so that CN
				// divided by DN is in the ratio the user has chosen CN_Original and DN_Original.
				// DN_Original and CN_Original are the CNF dimensions the user has set via console
				// input.

				if (USE_RANDOM_DN_MIN > DN_Original)
					// printf("warning: USE_RANDOM_DN_MIN > DN_Original (it is recommended to decrease USE_RANDOM_DN_MIN)!\n");
					USE_RANDOM_DN_MIN = DN_Original;
				if (USE_RANDOM_DN_MIN < 3)
					printf("warning: USE_RANDOM_DN_MIN < 3 (it is recommended to increase USE_RANDOM_DN_MIN)!\n");

				DN = (int)(
						(double)USE_RANDOM_DN_MIN + // DN is at least USE_RANDOM_DN_MIN
						(double)Rand() * ((double)DN_Original - (double)USE_RANDOM_DN_MIN + 1.0) / // DN is maximal DN_Original; '+ 1.0' as 'Rand() * X / RAND_MAX_PLUS_ONE' is maximal X-1
						(double)RAND_MAX_PLUS_ONE); // [required as Rand()'s output goes from 0 to 32767]

				CN = (int)(
						(double)CN_Original * (double)DN / (double)DN_Original * 1.0 / 2.0 + // CN goes from at least one half of CN_Original, adapted to ratio (DN/DN_Original)
						(double)Rand() * ((double)CN_Original * (double)DN / (double)DN_Original * 3.0 / 2.0) / // CN is maximal three half plus one half of CN_Original, adapted to ratio (DN/DN_Original)
						(double)RAND_MAX_PLUS_ONE); // [required as Rand()'s output goes from 0 to 32767]

				// some verification...
				if (DN < 3)
					DN = 3;
				if (DN > DIGIT_NUMBER_MAX)
					DN = DIGIT_NUMBER_MAX;

				if (CN < 1)
					CN = 1;
				if (CN > POSSIBLE_CLAUSE_NUMBER_MAX)
					CN = POSSIBLE_CLAUSE_NUMBER_MAX;
			}

			bool SolvableExponential;
			bool SolvablePolynomial;

			// generate a SAT CNF, or read it from a file (if TestMode == 2)
			switch (TestMode)
			{
			case TEST_MODE_RANDOM_CNF:

				CNF_CreateRandomly(DN, CN, CNF); // uses the random number generator
				break;

			case TEST_MODE_FILES:

				char FilePath[1024];

				SPRINTF1(FilePath, (const char*)FilePathFormat, TestRunIndex);

				if (CNF_ReadFromFile(DN, CN, CNF, FilePath) == false) // does NOT use the random number generator
					TestMode = TEST_MODE_EXIT;

				break;

			case TEST_MODE_RANDOM_INVERTED_CNF:

				CNF_CreateRandomlyWithInvertedClauses(DN, CN, CNF); // uses the random number generator
				break;

			case TEST_MODE_COMPLETE_CNF:

				CNF_CreateRandomlyWithAllPossibleClauses(DN, CN, CNF); // uses the random number generator
				break;

			case TEST_MODE_PIGEON_2:

				DN = 7;
				CN = 13;

				memcpy(CNF,
					"11----1"
					"--11--1"
					"----111"
					"0-0---1"
					"0---0-1"
					"--0-0-1"
					"-0-0--1"
					"-0---01"
					"---0-01"
					"11----0"
					"01----0"
					"10----0"
					"00----0",
					DN * CN * sizeof(char));

				break;

			case TEST_MODE_PIGEON_3:

				DN = 13;
				CN = 26;

				memcpy(CNF,
					"111----------"
					"---111-------"
					"------111----"
					"---------111-"
					"0--0--------1"
					"0-----0-----1"
					"0--------0--1"
					"---0--0-----1"
					"---0-----0--1"
					"------0--0--1"
					"-0--0-------1"
					"-0-----0----1"
					"-0--------0-1"
					"----0--0----1"
					"----0-----0-1"
					"-------0--0-1"
					"--0--0------1"
					"--0-----0---1"
					"--0--------01"
					"-----0--0---1"
					"-----0-----01"
					"--------0--01"
					"11----------0"
					"10----------0"
					"01----------0"
					"00----------0",
					DN * CN * sizeof(char));

				break;

			case TEST_MODE_ALL_COMPLETE:

				// first, verify PossibleClauses[][] and PossibleClauseNumber are filled properly

				DigitNumber = DN; // set global variable declared in Polynomial_Solver.h
				ClauseNumber = CN; // set global variable declared in Polynomial_Solver.h

				if (DigitNumberOld != DigitNumber ||
					SAT_TYPE_OLD != SAT_TYPE) // at first run or when DigitNumber or/and SAT_TYPE changed
				{
					Verify_PossibleClauses_And_Caches_Are_Initialized(); // initialize PossibleClauses[][] and PossibleClauseNumber
					TestMode_7_PossibleClausesOnOffCombinationIndex = 1;
				}
				else
				{
					TestMode_7_PossibleClausesOnOffCombinationIndex ++;
					if (TestMode_7_PossibleClausesOnOffCombinationIndex >= pow(2.0, (double)PossibleClauseNumber))
						TestMode = TEST_MODE_EXIT;
				}

				// then, build the CNF to be solved by using a special on/off combination of PossibleClauses[][] elements

				// IMPORTANT NOTE: the following works only if PossibleClauseNumber < 32, we'd have to
				// use 'u64 TestMode_7_PossibleClausesOnOffCombinationIndex' for greater values.

				CN = 0;
				for (int i = 0; i < PossibleClauseNumber; i ++)
				{
					if (TestMode_7_PossibleClausesOnOffCombinationIndex & (1 << i))
					{
						for (int m = 0; m < DN; m ++)
							CNF[CN * DigitNumber + m] = PossibleClauses[i][m];
						CN ++;
					}
				}
				break;

			case TEST_MODE_SOLVER_C_KILLER_CNF:

				CNF_CreateSolverCKillerCNF(DN, CN, CNF); // does NOT use the random number generator
				break;

			case TEST_MODE_PLANTED_CNF:

				CNF_CreatePlanted(DN, CN, CNF); // uses the random number generator
				break;

			default:

				printf("internal error: TestMode unknown !\n");
				break;
			}

			if (TestRunIndex < TEST_RUN_INDEX_START) // we have created the CNFs although we didn't solve them, this is to get re-producible tests because the Rand() function calls within the CNF creation functions are done in the same order, no matter what the value of TEST_RUN_INDEX_START is
				continue;

			// if a file could not be read:
			if (TestMode == TEST_MODE_EXIT)
				break;

			// show CNF info
			printf("DigitNumber=%d x ClauseNumber=%d; TestMode=%d, TestRunNumber=%d, %d-SAT\n", DN, CN, TestMode, TestRunNumber, SAT_TYPE);
			printf("\n");

			// show CNF if desired
			if (ShowCNFFlag)
			{
				printf("CNF to be solved:\n");
				CNF_Show(DN, CN, CNF);
			}

			// do the sail-safe, exponential solving
			if (DN <= 25) // exponential solver is too slow for greater DN values
			{
#ifndef MIN_OUTPUT
				printf("\n");
				printf("1) Using exponential algorithm to check for solvability...\n");
				printf("\n");

				if ((SolvableExponential = CNF_Solve_Exponential(DN, CN, CNF)) == true) // the exponential (O(2^n)) algorithm (guaranteed to work)
					printf(" * Exponential solution: solvable          (fail-safe)\n");
				else
					printf(" * Exponential solution: UNSAT             (fail-safe)\n");
#else
				SolvableExponential = CNF_Solve_Exponential(DN, CN, CNF);
#endif
			}
			else
			{
				SolvableExponential = false; // dummy value; should not be used if going here, must be set to avoid G++ (Linux) outputs a warning about non-initialized variable
			}

			// do the polynomial solving
// DoPolynomialSolving:;

			DWORD	t = GetTickCount();
			bool	PolynomialSolvingAbortedFlag = false;

#ifndef MIN_OUTPUT
			printf("\n");
			printf("2) Using polynomial algorithm to check for solvability...\n");
			printf("\n");
#endif
			if ((SolvablePolynomial = CNF_Solve_Polynomial(DN, CN, CNF, PolynomialSolvingAbortedFlag)) == true) // the polynomial algorithm (supposed to work ;-P )
			{
#ifndef MIN_OUTPUT
				printf("\n");
				t = (GetTickCount() - t + 500) / 1000; // +500 ms to round correctly
				printf(" * Polynomial solution : solvable          (must match exp result) (%d sec.)\n", t);
#endif
				SolvableNumber ++;
			}
			else
			if (PolynomialSolvingAbortedFlag)
			{
				printf("\n");
				t = (GetTickCount() - t + 500) / 1000; // +500 ms to round correctly
				printf(" * Polynomial solution : ABORTED           (solvability unknown) (%d sec.)\n", t);
			}
			else
			{
#ifndef MIN_OUTPUT
				printf("\n");
				t = (GetTickCount() - t + 500) / 1000; // +500 ms to round correctly
				printf(" * Polynomial solution : UNSAT             (must match exp result) (%d sec.)\n", t);
#endif
				UNSATNumber ++;
			}

			// compare exponential and polynomial result
			if (PolynomialSolvingAbortedFlag)
			{
				printf("\n=> ????? :/\n\n");

				UnknownNumber ++;
			}
			else if (DN > 25 /* && TestMode != TEST_MODE_FILES */ && TestMode != TEST_MODE_COMPLETE_CNF)
			{
				printf("\n=> ????? :/\n\n");

				UnknownNumber ++;
			}
			else if (SolvableExponential == SolvablePolynomial)
			{
				printf("\n=> Polynomial algorithm worked :)\n\n");

				SuccessNumber ++;
			}
			else
			{
				printf("\n=> Polynomial algorithm possibly failed :(\n\n"); // happened most times because there were programming errors, not algorithm design problems ;-P

				getchar();

				ErrorNumber ++;

				// goto DoPolynomialSolving; // try again?
			}

			if (TestMode == TEST_MODE_ALL_COMPLETE)
			{
				if (TestMode_7_PossibleClausesOnOffCombinationIndex >= pow(2.0, (double)PossibleClauseNumber) - 1)
					break;
			}

			if (_kbhit() != 0)
			{
				printf("abort? (y/n) ");
				GETS(S);
				if (S[0] == 'y')
					break;
				else
					printf("\n");
			}
		}

		printf("\n");
		printf("++++++++++++++++++++\n");
		printf("+++ End of Test. +++\n");
		printf("++++++++++++++++++++\n");
		printf("\n");
		printf("Success=%d (%d Solvable, %d UNSAT)   Error=%d   Unknown=%d\n", SuccessNumber, SolvableNumber, UNSATNumber, ErrorNumber, UnknownNumber);

		if (SuccessNumber > 0 && ErrorNumber == 0 && UnknownNumber == 0)
			printf("\n:) :) :) :) :)\n");

		printf("\n");

		printf("Press Return to finish test...\n");
		getchar();
	}

	return 0;
}

// ********************** END OF USER CONSOLE INTERFACE ***********************
