// (c)2012-2015 by Louis Coder (Matthias Mueller).

// Contact: visual_basic_louis@yahoo.com; Visit www.louis-coder.com!

// This is a helper program of the project 'Polynomial 3-SAT Solver'
// by Matthias Mueller. The program prints the 'Clause Table' (CT)
// to both the console and to a file named 'ClauseTable.txt' in the
// current directory of the executable (or the source code, if being
// run in Visual Studio).
// 
// The file is overwritten each time you print a new Clause Table.
// The file is meant to be opened with e.g. Notepad and the user
// shall examine the structure of the CT for a better understanding
// of the polynomial algorithm.
// 
// The Clause Table is used to find out if some 3-SAT CNF is solvable
// or not. For further information, please read the paper 'Polynomial
// Exact-3-SAT-Solving Algorithm' at:
// 
// "http://www.louis-coder.com/Polynomial_3-SAT_Solver/Polynomial_3-SAT_Solver.htm"

// under Windows, the following suppresses warnings
// about unsafe charstring operations (gets() etc.)
#define _CRT_SECURE_NO_DEPRECATE

#ifdef _WIN32 // should also be defined on Windows x64! (see MSDN)
	#include <stdio.h>			// fgets
	#include <string.h>			// strlen
	#include <tchar.h>			// _TCHAR
	#include <windows.h>		// GetTickCount

	#define GETS(s)				gets_s(s, sizeof(s) - 1)
#else
	#include <stdio.h>			// fgets
	#include <string.h>			// strlen
	#include <stdlib.h>			// atoi

	#define GETS(s)				if (fgets(s, sizeof(s) - 1, stdin)) {} // we need some useless return value check to prevent G++ (Linux) from outputting a warning
#endif

/* // currently not in use
// *** RANDOM ***
// NOTE: the Windows-function rand() is rather slow as
// it determines some process id every time it is called.
// So we copied the original Windows-rand() implementation's
// code and built our own random generator out of it!

#ifndef RAND_MAX
#define RAND_MAX	32767
#endif

// IMPORTANT: this random number generator will always
// return the same number succession. That means when
// you re-start the solver and do exactly the same as
// at last start, also the very same CNFs will be
// created. So if the solver should ever fail, please
// write down how you used the solver (test mode,
// DigitNumber, ClauseNumber etc.) to reproduce the error.
// Besides that, it is not ingenious to re-start the
// solver and use the same DigitNumber/ClauseNumber etc.
// combination, as the same CNFs will be created and
// tested again (we already know that they work).

int RandValue = 0;

void SRand(int Seed)
{ // initializes the random generator

	// NOTE: as long as you pass the same value to this function
	// (or void its calling), the Rand() function will return the
	// same succession of ClauseNumbers. That means in every run of
	// this program, results are re-producible!

	RandValue = Seed;
}

int Rand()
{ // returns a 'random' number between (both including) 0 and 32767

	return (((RandValue = RandValue * 214013L
            + 2531011L) >> 16) & 0x7fff); // copied from original rand() code
}

// *** END OF RANDOM ***
*/

// **************************************************************

#define	DIGIT_NUMBER_MIN	3
#define DIGIT_NUMBER_MAX	8 // more is possible, but the related CT would then be enormous; pay attention to array bounds, we might need a further element for null terminator!

#include "CTDisplayer.h"

#ifdef _WIN32
int _tmain(int argc, _TCHAR* argv[]) // the program starts here
#else
int main(int argc, char* argv[]) // the program starts here
#endif
{
	// set default values
	DigitNumber = 4; // defined in CTDisplayerAndTester.h
	SAT_TYPE = 3; // defined in CTDisplayerAndTester.h

#ifdef _WIN32
	// enlargen the console buffer
	COORD ConsoleDisplaySize;

	ConsoleDisplaySize.X = 896;
	ConsoleDisplaySize.Y = 896;

	SetConsoleScreenBufferSize(GetStdHandle(STD_OUTPUT_HANDLE), ConsoleDisplaySize);
#endif

	// do the main actions (displaying the CT)
	while (true)
	{
#ifdef _WIN32
		SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_INTENSITY);
#endif
		printf("*************************************************\n");
		printf("MAIN MENU ***************************************\n");
		printf("*************************************************\n");
		printf("\n");
		printf("d           : change DigitNumber (literal range)   Currently: %d\n", DigitNumber);
		printf("s           : change SAT Type (2-SAT or 3-SAT)     Currently: %d\n", SAT_TYPE);
		printf("[just Enter]: show Clause Table\n");

#ifdef _WIN32
		SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE);
#endif
		char Input[1024];

		GETS(Input);

		if (Input[0] == 'd')
		{
			while (true)
			{
				printf("Please enter new DigitNumber (between including %d and %d): ", DIGIT_NUMBER_MIN, DIGIT_NUMBER_MAX);

				GETS(Input);

				if (Input[0] == '\0')
					DigitNumber = 4; // default
				else
					DigitNumber = atoi(Input);

				if (DigitNumber >= DIGIT_NUMBER_MIN && DigitNumber <= DIGIT_NUMBER_MAX)
					break;
			}
		}
		else
		if (Input[0] == 's')
		{
			while (true)
			{
				printf("Please enter new SAT Type (either 2 or 3): ");

				GETS(Input);

				if (Input[0] == '\0')
					SAT_TYPE = 3;
				else
					SAT_TYPE = atoi(Input);

				if (SAT_TYPE == 2 || SAT_TYPE == 3)
					break;
			}
		}
		else
		if (Input[0] == '\0' || Input[0] == '\r' || Input[0] == '\n')
		{
			ClauseTable_Display();
		}
	}

	return 0;
}
