// (c)2016 by Louis Coder (real name Matthias Mueller).

// Refers to the paper 'Polynomial Exact-3-SAT-Solving Algorithm'
// by Matthias Mueller, version D-1.2. Paper can be downloaded on
// www.louis-coder.com or www.vixra.org/author/matthias_mueller.

// This is the 'Final Proof verification program' mentioned in
// the paper. This is *NOT* the actual polynomial solver!!!

/*
The present  program shall  test the  following supposition:   We
have a randomly  created JBase  and a KBase which  consists  each
of three '0' and the rest '-' chars. Example:

JBase = 000------
KBase = ---000---

JBase and  KBase  each have  DigitNumber   many chars  in  total.
DigitNumber   is for  each  test  run  chosen  randomly   between
(both including) 3 and DIGIT_NUMBER_MAX.

Now we do  a call to  PracticalRecursion().   There  the  program
adds a 'c'  char  at the  next location  where  JBase  AND  KBase
both have a '-' char. Example:

JK    = 000000c--
JBase = 000------
KBase = ---000---

Then the program uses  the ThreeOutOf() function  on JK with  the
'c' char added to get  all 3-SAT clauses  being CONTAINED  within
JK+'c'. ThreeOutOf() is here the C procedure that implements  the
functionality  of the 'tau()'  operator introduced  in the  paper
'Polynomial Exact-3-SAT-Solving Algorithm' version D-1.0, version
D-1.1 and version D-1.2.

At next,  the  code  checks  which  tuples  being  built  out  of
tau(JK+'c')   and tau(JK+'c')   did  appear  at  a special   code
location (see "disable  all tuples I,H  which were built  of...")
within PracticalRecursion()    and marks  these tuples  as  'have
been found':

I     = tau(000000c--)
H     = tau(000000c--)
JBase =     000------
KBase =     ---000---

The marking as  'have been found'  works like  this: The  program
did, in main()/_tmain(), mark  all EXPECTED I/H tuples as  'still
to be found'. Any of  those tuples is unmarked  when it has  been
found within PracticalRecursion().  At  the end of the  recursive
calls of  PracticalRecursion(),   ALL  tuples  must be  unmarked,
otherwise the program  would stop and  display an error  message.

After having unmarked tuples in the current call of
PracticalRecursion(),   the program  does  two recursion   steps.
Once tau(000000c--) becomes  the new J and the old JBase  becomes
the new K.

In the second  recursion  step,  which is  independent  from  the
first one, tau(000000c--)  becomes  the new  J and the old  KBase
(*K*Base!) becomes the new K.

In each of  the recursive  sub  calls, the  same 'c'  adding  and
tuple unmarking process is done again.

For instance, in  the (J=tau(000000c--),K=JBase)   sub call,  the
program unmarks all tuples build-able out of
(tau(000000cc-),tau(000000cc-)):

I = tau(000000cc-)
H = tau(000000cc-)
J = tau(000000c--)
K =     000------


THE PURPOSE OF THIS PROCEEDING IS THE FOLLOWING:

-the program  tests  if  really  all expected   I/H tuples   were
 processed  by the  recursive  code;  For details,   PLEASE  READ
 the 'Why UNSAT Detection  is Reliable'  section in the  document
 'Polynomial_3-SAT_Solver.pdf'  (should be within this  zip-file,
 in the directory  "Algorithm  D (newest version  - never  failed
 in tests)\Documents\".  ATTENTION: there  are several  documents
 with equal name  in the zip file,  this program  only refers  to
 the pdf in the  just mentioned  directory!  You know you  opened
 the right  pdf when  it's titled  'Version  D-1.2'  on the  very
 first page).

-as mentioned in  the pdf document,  we want to  verify that  all
 clause-combinations  from  the clause sets  S_x, x in  {1,2,...}
 were regarded by this test program,  and therewith by the  poly-
 nomial solver.  S was built by  using the tau  operator on  each
 Underlying Solution  of the Clause Table  Lines which  originate
 from 'adding all 0/1 combinations around JBase, KBase'. Here  in
 this test program, we do not  add both '0' and '1', but one  'c'
 to save CPU time and memory for storing which tuple has  already
 been found. The code  does not distinguish  between 'c' and  '0'
 and '-',  but between  '-'  and non-'-'  only,  to save  memory.
 Another unpublished  test  program written  by me  checked if  a
 variation of this code here with an explicit distinction between
 '0' and 'c' does fail, it didn't.  So I strongly assume  there's
 no problem with  this 'mixing' of  'c' and '0'. The  unpublished
 test program  that  handled '0'  and 'c'  separately  could  not
 be run with  larger DigitNumbers   (not larger  than 8),  that's
 the problem.

-the explicit  recursive  calls  of PracticalRecursion()    shall
 SIMULATE the work  of the Practical Recursion  within the  poly-
 nomial 3-SAT solver. This means  that the test program HERE  has
 exponential  complexity,  but this  does not matter  as we  only
 want to know if all I/H combinations were found; The  polynomial
 solver does NOT use this recursive procedure calling.

If you have questions, mail louis@louis-coder.com.

So, just  run  this  program;   if  it does   NOT stop  with   an
error  message,  everything's    fine and  we  can still   assume
the polynomial   3-SAT solving  algorithm   by Matthias   Mueller
is not obviously faulty.
*/

#ifdef _WIN32
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <windows.h>
#include <math.h>
#else
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/times.h>
#include <math.h>
#endif

#define DIGIT_NUMBER_MAX				10				// 'DigitNumber' is another expression for 'n' (both introduced in the paper 'Polynomial Exact-3-SAT-Solving Algorithm')
#define POSSIBLE_CLAUSE_NUMBER_MAX		1000			// As a rule of thumb, DIGIT_NUMBER_MAX^3 is enough

int		DigitNumber = 0;								// The current DigitNumber (randomized before every test)
int		TupleToBeFoundNumberLeft = 0;					// How many tuples need still to be found

char	JBase[DIGIT_NUMBER_MAX + 1];					// The J at very bottom of ClauseStack
char	KBase[DIGIT_NUMBER_MAX + 1];					// The K at very bottom of ClauseStack

int		Place01Pos[DIGIT_NUMBER_MAX + 1];				// The next char index where JBase and KBase are both '-'

#define TUPLE_TO_BE_FOUND_NUMBER_MAX	1048576			// 2^(DIGIT_NUMBER_MAX * 2)

bool	TupleToBeFound[TUPLE_TO_BE_FOUND_NUMBER_MAX];	// true if a tuple identified by the array index is still to be found, false if tuple needn't to be found or has been found

// The 'tau' operator:
void ThreeOutOf(char* InClause, int& OutNumber, char OutArray[POSSIBLE_CLAUSE_NUMBER_MAX][DIGIT_NUMBER_MAX + 1])
{ // write, to OutArray, all combinations of taking exactly three '0' or/and '1' from InClause, in all possible combinations

	OutNumber = 0;
	for (int c1 = 0; c1 < DigitNumber - 2; c1 ++)
	{
		if (InClause[c1] == '-')
			continue;

		for (int c2 = c1 + 1; c2 < DigitNumber - 1; c2 ++)
		{
			if (InClause[c2] == '-')
				continue;

			for (int c3 = c2 + 1; c3 < DigitNumber - 0; c3 ++)
			{
				if (InClause[c3] == '-')
					continue;

				if (OutNumber >= POSSIBLE_CLAUSE_NUMBER_MAX)
				{
					printf("Too great OutNumber!\n");
					getchar();
				}
				for (int m = 0; m < DigitNumber; m ++)
					OutArray[OutNumber][m] = '-';
				OutArray[OutNumber][c1] = InClause[c1];
				OutArray[OutNumber][c2] = InClause[c2];
				OutArray[OutNumber][c3] = InClause[c3];
				OutArray[OutNumber][DigitNumber] = '\0';
				OutNumber ++;
			}
		}
	}
}

// Get an array index which is unique for each C1, C2 tuple:
int GetTupleToBeFoundIndex(char* C1, char* C2)
{
	// Works up to DigitNumber = 15, but not beyond,
	// as then an int has not enough bits to set.

	int TupleToBeFoundIndex = 0; // 0 means no bits set

	for (int m = 0; m < DigitNumber; m ++)
		if (C1[m] != '-')
			TupleToBeFoundIndex |= (1 << m); // set bit #m

	for (int m = DigitNumber; m < DigitNumber * 2; m ++)
		if (C2[m - DigitNumber] != '-')
			TupleToBeFoundIndex |= (1 << m); // set bit #(DigitNumber + m)

	return TupleToBeFoundIndex;
}

void TupleToBeFound_SetAllNotFound()
{
	memset(TupleToBeFound, 0, TUPLE_TO_BE_FOUND_NUMBER_MAX * sizeof(bool)); // set whole array to 'false'
}

bool TupleToBeFound_SetToBeFound(char* C1, char* C2)
{ // returns previous value

	int TupleToBeFoundIndex = GetTupleToBeFoundIndex(C1, C2);
	bool OldValue = TupleToBeFound[TupleToBeFoundIndex];
	TupleToBeFound[TupleToBeFoundIndex] = true;
	return OldValue;
}

bool TupleToBeFound_SetHasBeenFound(char* C1, char* C2)
{ // returns previous value

	int TupleToBeFoundIndex = GetTupleToBeFoundIndex(C1, C2);
	bool OldValue = TupleToBeFound[TupleToBeFoundIndex];
	TupleToBeFound[TupleToBeFoundIndex] = false;
	return OldValue;
}

void TupleToBeFound_VerifyAllFound()
{
	int TupleToBeFoundNumber = (int)pow(2.0, (double)(2 * DigitNumber)); // (2^DigitNumber values for C1) * (2^DigitNumber values for C2)
	if (TupleToBeFoundNumber < 0 ||
		TupleToBeFoundNumber > TUPLE_TO_BE_FOUND_NUMBER_MAX)
	{
		printf("warning in TupleToBeFound_VerifyAllFound: TupleToBeFoundNumber invalid!\n");
		getchar();
	}

	bool Error = false;
	for (int i = 0; i < TupleToBeFoundNumber; i ++)
		if (((bool*)TupleToBeFound)[i] != 0) // 0 means 'false'
		{
			printf("Error: Tuple #%d still to be found!\n", i);
			Error = true;
		}

	if (Error)
		getchar();

	// if no error message was shown, all is ok (all flags unset)
}

bool PracticalRecursion(char* J, char* K, int Place01PosIndex)
{
	// get the overlay of J and K
	// Example:
	// J  = 000------
	// K  = --0-00---
	// JK = 000-00---
	// 
	char	JK[DIGIT_NUMBER_MAX + 1];

	for (int m = 0; m <= DigitNumber; m ++) // copy also null terminator
		if (J[m] != '-')
			JK[m] = J[m];
		else
			JK[m] = K[m];

	// Place 'c' as placeholder for '0' and '1' (to save CPU time and memory
	// we do not check 0 and 1 paths (I and H) but simulate checking the 0 or
	// 1 path only (c can be seen as 0 or alternatively 1)):
	// 
	if (Place01Pos[Place01PosIndex] >= 0 &&
		JK[Place01Pos[Place01PosIndex]] == '-')
	{
		JK[Place01Pos[Place01PosIndex]] = 'c'; // simulates adding the 'conflict digits' of I and H at position p
	}

	int		INewNumber = 0;
	char	INewArray[POSSIBLE_CLAUSE_NUMBER_MAX][DIGIT_NUMBER_MAX + 1];

	// use the tau operator on JK+'c'
	ThreeOutOf(JK, INewNumber, INewArray);

	// Disable all tuples I,H which were built of
	// I = tau(JK+'c'),
	// H = tau(JK+'c'):
	// 
	for (int i = 0; i < INewNumber; i ++)
	{
		for (int h = 0; h < INewNumber; h ++) // I and H need not to be equal!
		{
			if (TupleToBeFound_SetHasBeenFound(INewArray[i], INewArray[h]) == true)
			{
				TupleToBeFoundNumberLeft --;

				if (TupleToBeFoundNumberLeft == 0) // leave recursion if all flags unset (optimization to speed-up)
					return true;
			}
		}
	}

	// Do the recursion, which means add further 'c' digits in
	// deeper recursion layers and there mark tuples as found:
	// 
	if (Place01Pos[Place01PosIndex + 0] >= 0 &&
		Place01Pos[Place01PosIndex + 1] >= 0) // have we not yet placed as many 'c' chars as possible?
	{
		for (int i = 0; i < INewNumber; i ++)
		{
			/*
			// The behavior of the following code lines from the original
			// polynomial SAT solver shall here be 'simulated' by doing
			// an explicit recursive procedure calling:

			if ((!ActiveClauseTuple[I][J] || !ActiveClauseTuple[I][K]) &&
			// we go deeper into ^ this and ...              ^ this path!
				(!ActiveClauseTuple[H][J] || !ActiveClauseTuple[H][K]))
			{
				if (Does_jk_Contain_ih(I, H, J, K))
				{
					ActiveClauseTuple[J][K] = false; // this code line disables tuple (J,K)
				}
			}
			*/

			if (PracticalRecursion(INewArray[i], J, Place01PosIndex + 1))
				return true;

			if (PracticalRecursion(INewArray[i], K, Place01PosIndex + 1))
				return true;
		}
	}

	return false;
}

#ifdef _WIN32
int _tmain(int argc, _TCHAR* argv[])
#else
int main(int argc, char* argv[])
#endif
{
#ifdef _WIN32
typedef __int64 int64_t;
#define RAND_MAX_PLUS_ONE ((int64_t)RAND_MAX + (int64_t)1)
	int RandomSeed = (GetTickCount() % 1024); // use randomized seed at each program start up
#else
#define RAND_MAX_PLUS_ONE ((int64_t)RAND_MAX + (int64_t)1)
	tms tm;
	int RandomSeed = (times(&tm) % 1024); // use randomized seed at each program start up
#endif

	srand(RandomSeed);

	TupleToBeFound_SetAllNotFound();

	for (int Iteration = 1; Iteration <= 100000000; Iteration ++)
	{
		printf("****** NEW TEST #%d   (RandomSeed=%d)\n", Iteration, RandomSeed);

		// set DigitNumber randomly
		int DigitNumberMax = DIGIT_NUMBER_MAX;

		DigitNumber = (int)((int64_t)3 + (int64_t)rand() * (int64_t)(DigitNumberMax - 3 + 1) / (int64_t)(RAND_MAX_PLUS_ONE));

		// define basis J and K randomly
		for (int m = 0; m < DigitNumber; m ++)
			JBase[m] = '-';
		JBase[DigitNumber] = '\0';
		for (int m = 0; m < DigitNumber; m ++)
			KBase[m] = '-';
		KBase[DigitNumber] = '\0';

		for (int m = 0; m < 3; m ++)
		{
			int Place0Pos = (int)((int64_t)rand() * (int64_t)DigitNumber / (int64_t)(RAND_MAX_PLUS_ONE));
			if (JBase[Place0Pos] == '-')
				JBase[Place0Pos] = '0';
			else
				m --;
		}
		for (int m = 0; m < 3; m ++)
		{
			int Place0Pos = (int)((int64_t)rand() * (int64_t)DigitNumber / (int64_t)(RAND_MAX_PLUS_ONE));
			if (KBase[Place0Pos] == '-')
				KBase[Place0Pos] = '0';
			else
				m --;
		}

		// strcpy(JBase, "000-----"); // enable for debugging (then verify DigitNumber fits!)
		// strcpy(KBase, "---000--"); // enable for debugging (then verify DigitNumber fits!)

		printf("DigitNumber=%d\n", DigitNumber);
		printf("JBase=%s\n", JBase);
		printf("KBase=%s\n", KBase);

		// verify there's no misconfiguration
		if (DigitNumber > DIGIT_NUMBER_MAX)
		{
			printf("warning: DigitNumber too large!\n");
			getchar();
		}
		if (DigitNumber < 3)
		{
			printf("warning: DigitNumber too small!\n");
			getchar();
		}
		if (JBase[DigitNumber] != '\0' ||
			KBase[DigitNumber] != '\0')
		{
			printf("warning: DigitNumber seems not to fit to content of JBase or/and KBase!\n");
			getchar();
		}
		if (DigitNumber * DigitNumber > TUPLE_TO_BE_FOUND_NUMBER_MAX)
		{
			printf("warning: DigitNumber exceeds TUPLE_TO_BE_FOUND_NUMBER_MAX!\n");
			getchar();
		}
		if ((int)pow(2.0, (double)(DigitNumber * 2)) > TUPLE_TO_BE_FOUND_NUMBER_MAX)
		{
			printf("warning: DigitNumber too large compared to TUPLE_TO_BE_FOUND_NUMBER_MAX!\n");
			getchar();
		}

		// mark tuples as 'still to be found'
		char	PossibleSolution[DIGIT_NUMBER_MAX + 1];

		for (int m = 0; m < DigitNumber; m ++)
			if (JBase[m] == '0' || KBase[m] == '0')
				PossibleSolution[m] = '0';
			else
				PossibleSolution[m] = 'c';
		PossibleSolution[DigitNumber] = '\0';

		int		INewNumber = 0;
		char	INewArray[POSSIBLE_CLAUSE_NUMBER_MAX][DIGIT_NUMBER_MAX + 1];

		ThreeOutOf(PossibleSolution, INewNumber, INewArray);

		TupleToBeFoundNumberLeft = 0;

		for (int i = 0; i < INewNumber; i ++)
		{
			for (int h = 0; h < INewNumber; h ++)
			{
				TupleToBeFound_SetToBeFound(INewArray[i], INewArray[h]);
				TupleToBeFoundNumberLeft ++;
			}
		}

		// set order in which we'll place 'c'
		int Place01PosNumber = 0;
		for (int m = 0; m < DigitNumber; m ++)
			if (JBase[m] == '-' && KBase[m] == '-')
				Place01Pos[Place01PosNumber ++] = m;
		Place01Pos[Place01PosNumber] = -1; // end marker; NO increase of counter!

		// do the recursion
		printf("Doing actual recursion...\n");
		PracticalRecursion(JBase, KBase, 0);

		// check result of recursion
		printf("Finished recursion. Checking if all expected tuples were found...\n");
		TupleToBeFound_VerifyAllFound();
		printf("TupleToBeFoundNumberLeft=%d\n", TupleToBeFoundNumberLeft);
		if (TupleToBeFoundNumberLeft != 0)
		{
			printf("Error: TupleToBeFoundNumberLeft should be 0!\n");
			getchar();
		}

		printf("FINISHED current J,K test.\n");
	}

	return 0;
}
