// (c)2017 by Louis Coder (real name Matthias Mueller).

// Refers to the paper 'Polynomial Exact-3-SAT-Solving Algorithm'
// by Matthias Mueller, version D-1.3. Paper can be downloaded on
// www.louis-coder.com or www.vixra.org/author/matthias_mueller.

// This is the 'Final Proof Verification Program Two' (abbrev. FPVP Two)
// mentioned in the paper. This is *NOT* the actual polynomial solver!!!
// The actual polynomial solver is contained, besides this FPVP Two,
// within the zip file "Polynomial_3-SAT_Solver.zip", downloadable
// for free from www.louis-coder.com.

/*
In the document "Polynomial Exact-3-SAT-Solving Algorithm" it was claimed in the
"Why UNSAT Detection is Reliable" proof that in the Practical Recursion examina-
tion the following clauses are expected to come up:

I_Lambda   element of   \tau({d^*}ddddddd0{-^*}) <-- we check this if JOrI = 1
H_Lambda   element of   \tau({d^*}ddddddd1{-^*})
J_Lambda   element of   \tau({d^*}ddddddd-{-^*}) <-- we check this if JOrI = 0
K_Lambda   element of   \tau({d^*}dddddd--{-^*})

This program just checks if really EVERY expected clause from the J_Lambda or
I_Lambda sets comes up. \tau(C) is the set of 3-SAT clauses being contained within
an x-SAT clause C (x >= 3). The higher Lambda, the more elements the \tau() sets
have, because more Practical Recursion steps have been made. In each Practical
Recursion step, another '-' digit has been replaced by a '0' digit within the
participated clauses (please see this project's procedure 'PracticalRecursion()').

Because this test program is an adjusted Final Proof Verification Program (FPVP),
its code is similar to that of the FPVP. Some source code of the FPVP remained
commented-out here (e.g. in the procedure PracticalRecursion()) or is used in a
slightly different way than in the FPVP (for example TupleToBeFound[][] is here
just used to find clauses, not clause tuples). The intention behind is just to
save programming work (additional work could generate new undiscovered programming
faults) and to make the user quicker understand the program because it shares
already familiar code with the FPVP.

If the program does not stop with an error message but just keeps running, every-
thing's as expected. The program might hang some time when doing the Practical
Recursion, but in case of an error an explicit message will be shown.
In all ~30,000 test runs I did no error came up.
*/

#ifdef _WIN32
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <windows.h>
#include <math.h>
#else
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/times.h>
#include <math.h>
#endif

#define DIGIT_NUMBER_MAX				14				// 'DigitNumber' is another expression for 'n' (both introduced in the paper 'Polynomial Exact-3-SAT-Solving Algorithm')
#define POSSIBLE_CLAUSE_NUMBER_MAX		2800			// As a rule of thumb, DIGIT_NUMBER_MAX^3 is enough

int		DigitNumber = 0;								// The current DigitNumber (randomized before every test)
int		TupleToBeFoundNumberLeft = 0;					// How many tuples need still to be found

char	JBase[DIGIT_NUMBER_MAX + 1];					// The J at very bottom of ClauseStack
char	KBase[DIGIT_NUMBER_MAX + 1];					// The K at very bottom of ClauseStack

int		Place0Pos[DIGIT_NUMBER_MAX + 1];				// The next char index where JBase and KBase are both '-'

#define TUPLE_TO_BE_FOUND_NUMBER_MAX	256 * 1048576	// 2^(DIGIT_NUMBER_MAX * 2)

bool	TupleToBeFound[TUPLE_TO_BE_FOUND_NUMBER_MAX];	// flags of which each is true if a tuple identified by the array index is still to be found, false if tuple needn't to be found or has been found

// The 'tau' operator:
void ThreeOutOf(char* InClause, int& OutNumber, char OutArray[POSSIBLE_CLAUSE_NUMBER_MAX][DIGIT_NUMBER_MAX + 1])
{ // write, to OutArray, all combinations of taking exactly three '0' or/and '1' or/and 'd' from InClause, in all possible combinations

	OutNumber = 0;
	for (int c1 = 0; c1 < DigitNumber - 2; c1 ++)
	{
		if (InClause[c1] == '-')
			continue;

		for (int c2 = c1 + 1; c2 < DigitNumber - 1; c2 ++)
		{
			if (InClause[c2] == '-')
				continue;

			for (int c3 = c2 + 1; c3 < DigitNumber - 0; c3 ++)
			{
				if (InClause[c3] == '-')
					continue;

				if (OutNumber >= POSSIBLE_CLAUSE_NUMBER_MAX)
				{
					printf("Too great OutNumber!\n");
					getchar();
				}
				for (int m = 0; m < DigitNumber; m ++)
					OutArray[OutNumber][m] = '-';
				OutArray[OutNumber][c1] = InClause[c1];
				OutArray[OutNumber][c2] = InClause[c2];
				OutArray[OutNumber][c3] = InClause[c3];
				OutArray[OutNumber][DigitNumber] = '\0';
				OutNumber ++;
			}
		}
	}
}

// Get an array index which is unique for each C1, C2 tuple:
int GetTupleToBeFoundIndex(char* C1, char* C2)
{
	// Works up to DigitNumber = 15, but not beyond,
	// as then an int has not enough bits to set.

	int TupleToBeFoundIndex = 0; // 0 means no bits set

	for (int m = 0; m < DigitNumber; m ++)
		if (C1[m] != '-')
			TupleToBeFoundIndex |= (1 << m); // set bit #m

	for (int m = DigitNumber; m < DigitNumber * 2; m ++)
		if (C2[m - DigitNumber] != '-')
			TupleToBeFoundIndex |= (1 << m); // set bit #(DigitNumber + m)

	return TupleToBeFoundIndex;
}

void TupleToBeFound_SetAllNotFound()
{
	memset(TupleToBeFound, 0, TUPLE_TO_BE_FOUND_NUMBER_MAX * sizeof(bool)); // set whole array to 'false'
}

bool TupleToBeFound_SetToBeFound(char* C1, char* C2)
{ // returns previous value

	int TupleToBeFoundIndex = GetTupleToBeFoundIndex(C1, C2);
	bool OldValue = TupleToBeFound[TupleToBeFoundIndex];
	TupleToBeFound[TupleToBeFoundIndex] = true;
	return OldValue;
}

bool TupleToBeFound_SetHasBeenFound(char* C1, char* C2)
{ // returns previous value

	int TupleToBeFoundIndex = GetTupleToBeFoundIndex(C1, C2);
	bool OldValue = TupleToBeFound[TupleToBeFoundIndex];
	TupleToBeFound[TupleToBeFoundIndex] = false;
	return OldValue;
}

void TupleToBeFound_VerifyAllFound()
{
	int TupleToBeFoundNumber = (int)pow(2.0, (double)(2 * DigitNumber)); // (2^DigitNumber values for C1) * (2^DigitNumber values for C2)
	if (TupleToBeFoundNumber < 0 ||
		TupleToBeFoundNumber > TUPLE_TO_BE_FOUND_NUMBER_MAX)
	{
		printf("warning in TupleToBeFound_VerifyAllFound: TupleToBeFoundNumber invalid!\n");
		getchar();
	}

	bool Error = false;
	for (int i = 0; i < TupleToBeFoundNumber; i ++)
		if (((bool*)TupleToBeFound)[i] != 0) // 0 means 'false'
		{
			printf("Error: Tuple #%d still to be found!\n", i);
			Error = true;
		}

	if (Error)
		getchar();

	// if no error message was shown, all is ok (all flags unset)
}

// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// global numbers; are set to random values
// before each test iteration (see _tmain())
int Lambda = 2;
int JOrI = 0;
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

bool PracticalRecursion(char* J, char* K, int Place0PosIndex)
{
	// get the overlay of J and K
	// Example:
	// J  = ddd------
	// K  = --d-dd---
	// JK = ddd-dd---
	// 
	char	JK[DIGIT_NUMBER_MAX + 1];

	for (int m = 0; m <= DigitNumber; m ++) // copy also null terminator
		if (J[m] != '-')
			JK[m] = J[m];
		else
			JK[m] = K[m];

	// Place '0' at a position named 'p' in the paper (and then use the \tau
	// operator, which is here the procedure ThreeOutOf()) to generate I.
	// H would get a '1' placed, but the H branch is here not checked because
	// its processing/verification is, bar the placing of '1' instead of '0',
	// equal to that of I.
	// 
	if (Place0Pos[Place0PosIndex] >= 0 &&
		JK[Place0Pos[Place0PosIndex]] == '-')
	{
		JK[Place0Pos[Place0PosIndex]] = '0'; // simulates adding the 'conflict digit' of I at position p
	}

	int		INewNumber = 0;
	char	INewArray[POSSIBLE_CLAUSE_NUMBER_MAX][DIGIT_NUMBER_MAX + 1];

	// use the \tau operator on JK+'0'
	ThreeOutOf(JK, INewNumber, INewArray);

	/* // old FPVP (One) code:
	// Disable all tuples I,H which were built of
	// I = \tau(JK+'0'),
	// H = \tau(JK+'0'):
	// 
	for (int i = 0; i < INewNumber; i ++)
	{
		for (int h = 0; h < INewNumber; h ++) // I and H need not to be equal!
		{
			if (TupleToBeFound_SetHasBeenFound(INewArray[i], INewArray[h]) == true)
			{
				TupleToBeFoundNumberLeft --;

				if (TupleToBeFoundNumberLeft == 0) // leave recursion if all flags unset (optimization to speed-up)
					return true;
			}
		}
	}
	*/
	// new FPVP Two code:
	if (Place0PosIndex == Lambda) // are we in the right recursion depth to check for J or I?
	{
		// *******************************************
		// *** MARK HERE ALL CLAUSES WHICH COME UP ***
		// *******************************************
		// 
		// At the very end of the recursive PracticalRecursion() calling, all clauses
		// (resp. in practice tuples of two equal clauses) must have been marked as found.
		// In _tmain() (see below) we marked the clauses we EXPECT to come up as 'to be found'.
		// So finally EACH of those clauses must have been marked as 'has been found' by the
		// following code lines. The final check if all clauses have been found is done by
		// TupleToBeFound_VerifyAllFound(), which is called out of _tmain().
		// 
		if (JOrI == 0)
		{
			if (TupleToBeFound_SetHasBeenFound(J, J) == true) // mark J_Lambda as has been found
			{
				TupleToBeFoundNumberLeft --;

				if (TupleToBeFoundNumberLeft == 0) // leave recursion if all flags unset (optimization to speed-up)
					return true;
			}
		}
		else
		{
			for (int i = 0; i < INewNumber; i ++)
			{
				if (TupleToBeFound_SetHasBeenFound(INewArray[i], INewArray[i]) == true) // mark I_Lambda as has been found
				{
					TupleToBeFoundNumberLeft --;

					if (TupleToBeFoundNumberLeft == 0) // leave recursion if all flags unset (optimization to speed-up)
						return true;
				}
			}
		}
		// END OF MARKING CLAUSES
	}

	// Do the recursion, which means add further '0' digits in
	// deeper recursion layers and there mark tuples as found:
	// 
	if (Place0PosIndex > Lambda)
		return false;

	if (Place0Pos[Place0PosIndex + 0] >= 0 &&
		Place0Pos[Place0PosIndex + 1] >= 0) // have we not yet placed as many '0' chars as possible?
	{
		for (int i = 0; i < INewNumber; i ++)
		{
			/*
			// The behavior of the following code lines from the original
			// polynomial SAT solver shall here be 'simulated' by doing
			// an explicit recursive procedure calling:

			if ((!ActiveClauseTuple[I][J] || !ActiveClauseTuple[I][K]) &&
			// we go deeper into ^ this and ...              ^ this path!
				(!ActiveClauseTuple[H][J] || !ActiveClauseTuple[H][K]))
			{
				if (Does_jk_Contain_ih(I, H, J, K))
				{
					ActiveClauseTuple[J][K] = false; // this code line disables tuple (J,K)
				}
			}
			*/

			if (PracticalRecursion(INewArray[i], J, Place0PosIndex + 1))
				return true;

			// Enable the following to check at a deeper recursion level
			// only each (I,J) branch; this seems to work and speeds up
			// the program. There's no explicit proof if and why we can
			// save the (I,K) recursion but it works in practice.
			if (Place0PosIndex < 1) // enable to pass (I,K) on only to Lambda == 1
			{
				if (PracticalRecursion(INewArray[i], K, Place0PosIndex + 1))
					return true;
			}
		}
	}

	return false;
}

#ifdef _WIN32
int _tmain(int argc, _TCHAR* argv[])
#else
int main(int argc, char* argv[])
#endif
{
#ifdef _WIN32
typedef __int64 int64_t;
#define RAND_MAX_PLUS_ONE ((int64_t)RAND_MAX + (int64_t)1)
	int RandomSeed = (GetTickCount() % 1024); // use randomized seed at each program start up
#else
#define RAND_MAX_PLUS_ONE ((int64_t)RAND_MAX + (int64_t)1)
	tms tm;
	int RandomSeed = (times(&tm) % 1024); // use randomized seed at each program start up
#endif

	srand(RandomSeed);

	TupleToBeFound_SetAllNotFound();

	for (int Iteration = 1; Iteration <= 100000000; Iteration ++)
	{
		printf("****** NEW TEST #%d   (RandomSeed=%d)\n", Iteration, RandomSeed);

		int DigitNumberMax = DIGIT_NUMBER_MAX;

		// Set DigitNumber randomly; we use a minimal DigitNumber
		// (which is the clause length in Clause Line Notation) of 8
		// to avoid the problem JBase and KBase get chosen in a way
		// no Practical Recursion can be done any more; this could be
		// programmed more precise but would make the code more compli-
		// cated without much higher validity:
		DigitNumber = (int)((int64_t)8 + (int64_t)rand() * (int64_t)(DigitNumberMax - 8 + 1) / (int64_t)(RAND_MAX_PLUS_ONE));

		// define basis J and K randomly
		for (int m = 0; m < DigitNumber; m ++)
			JBase[m] = '-';
		JBase[DigitNumber] = '\0';
		for (int m = 0; m < DigitNumber; m ++)
			KBase[m] = '-';
		KBase[DigitNumber] = '\0';

		for (int m = 0; m < 3; m ++)
		{
			int Place0Pos = (int)((int64_t)rand() * (int64_t)DigitNumber / (int64_t)(RAND_MAX_PLUS_ONE));
			if (JBase[Place0Pos] == '-')
				JBase[Place0Pos] = 'd';
			else
				m --;
		}
		for (int m = 0; m < 3; m ++)
		{
			int Place0Pos = (int)((int64_t)rand() * (int64_t)DigitNumber / (int64_t)(RAND_MAX_PLUS_ONE));
			if (KBase[Place0Pos] == '-')
				KBase[Place0Pos] = 'd';
			else
				m --;
		}

		// decide if to test for completeness either
		// \tau(J_Lambda) or \tau(I_Lambda)
		if (rand() < (RAND_MAX_PLUS_ONE / 2))
			JOrI = 0;
		else
			JOrI = 1;

		// randomly select Lambda (Practical Recursion level where
		// \tau(J_Lambda) or \tau(I_Lambda) is to be checked for
		// completeness)
		int FreeNumber = 0;
		for (int m = 0; m < DigitNumber; m ++)
			if (JBase[m] == '-' && KBase[m] == '-')
				FreeNumber ++;

		// we cannot check for all J if we did not do a recursion step,
		// so begin not with less than a minimal Lambda == 1 if JOrI == 0:
		Lambda = (int)((JOrI == 0 ? 1 : 0) + (int64_t)(FreeNumber - (JOrI == 0 ? 1 : 0)) * (int64_t)rand() / (int64_t)(RAND_MAX_PLUS_ONE));

		// debug output
		printf("DigitNumber=%d\n", DigitNumber);
		printf("JBase=%s\n", JBase);
		printf("KBase=%s\n", KBase);

		// verify there's no misconfiguration
		if (DigitNumber > DIGIT_NUMBER_MAX)
		{
			printf("warning: DigitNumber too large!\n");
			getchar();
		}
		if (DigitNumber < 3)
		{
			printf("warning: DigitNumber too small!\n");
			getchar();
		}
		if (JBase[DigitNumber] != '\0' ||
			KBase[DigitNumber] != '\0')
		{
			printf("warning: DigitNumber seems not to fit to content of JBase or/and KBase!\n");
			getchar();
		}
		if (DigitNumber * DigitNumber > TUPLE_TO_BE_FOUND_NUMBER_MAX)
		{
			printf("warning: DigitNumber exceeds TUPLE_TO_BE_FOUND_NUMBER_MAX!\n");
			getchar();
		}
		if ((int)pow(2.0, (double)(DigitNumber * 2)) > TUPLE_TO_BE_FOUND_NUMBER_MAX)
		{
			printf("warning: DigitNumber too large compared to TUPLE_TO_BE_FOUND_NUMBER_MAX!\n");
			getchar();
		}

		// mark clauses as 'still to be found'
		char	PossibleSolution[DIGIT_NUMBER_MAX + 1];
		int		Placed0Number = 0;

		for (int m = 0; m < DigitNumber; m ++)
			if (JBase[m] == 'd' || KBase[m] == 'd')
				PossibleSolution[m] = 'd';
			else if ((Placed0Number ++) <= Lambda - 1 + JOrI)
				PossibleSolution[m] = '0';
			else
				PossibleSolution[m] = '-';

		PossibleSolution[DigitNumber] = '\0';

		int		INewNumber = 0;
		char	INewArray[POSSIBLE_CLAUSE_NUMBER_MAX][DIGIT_NUMBER_MAX + 1];

		ThreeOutOf(PossibleSolution, INewNumber, INewArray);

		TupleToBeFoundNumberLeft = 0;

		/* // FPVP (One) original code:
		for (int i = 0; i < INewNumber; i ++)
		{
			for (int h = 0; h < INewNumber; h ++)
			{
				TupleToBeFound_SetToBeFound(INewArray[i], INewArray[h]);
				TupleToBeFoundNumberLeft ++;
			}
		}
		*/
		// FPVP Two code:
		// NOTE: we actually check for single clauses only. As this code was taken from
		// the Final Proof Verification Program (One), which checks for tuples of clauses,
		// we just mark/unmark tuples built out of two of the same clauses.
		for (int i = 0; i < INewNumber; i ++)
		{
			if (JOrI == 0)
				printf("Need to find as J at Practical Recursion depth %d: %s\n", Lambda, INewArray[i]);
			else
				printf("Need to find as I at Practical Recursion depth %d: %s\n", Lambda, INewArray[i]);
			TupleToBeFound_SetToBeFound(INewArray[i], INewArray[i]); // i and i
			TupleToBeFoundNumberLeft ++;
		}
		printf("Iteration %d: Lambda = %d, JOrI = %d\n", Iteration, Lambda, JOrI);
		// getchar();

		// set order in which we'll place '0'
		int Place01PosNumber = 0;
		for (int m = 0; m < DigitNumber; m ++)
			if (JBase[m] == '-' && KBase[m] == '-')
				Place0Pos[Place01PosNumber ++] = m;
		Place0Pos[Place01PosNumber] = -1; // end marker; NO increase of counter!

		// do the recursion
		printf("Doing actual recursion...\n");
		PracticalRecursion(JBase, KBase, 0);

		// check result of recursion
		printf("Finished recursion. Checking if all expected tuples were found...\n");
		TupleToBeFound_VerifyAllFound();
		printf("TupleToBeFoundNumberLeft=%d\n", TupleToBeFoundNumberLeft);
		if (TupleToBeFoundNumberLeft != 0)
		{
			printf("Error: TupleToBeFoundNumberLeft should be 0!\n"); // Error: not all expected clauses were found!
			getchar();
		}

		printf("FINISHED current J,K test.\n");
	}

	return 0;
}
